########################################################################################################################
# Module: ssm/linear_gaussian/linear_gaussian.py
# Description: Linear TemporalGaussian state-space models
#
# Web: https://github.com/SamDuffield/mocat
########################################################################################################################


from typing import Union, Any

from jax import numpy as jnp, random
from mocat.src.utils import gaussian_potential, extract_dimension, reset_covariance
from mocat.src.ssm.ssm import StateSpaceModel


class LinearGaussian(StateSpaceModel):
    name = 'Linear Gaussian'

    # x_{t_new} ~ N(F_{t_new} @ x_{t_previous}, Q_{t_new})
    # y_t ~ N(H_t @ x_t, R_t)

    def get_initial_mean(self,
                         t: Union[float, None]) -> jnp.ndarray:
        raise NotImplementedError

    def get_initial_covariance_sqrt(self,
                                    t: Union[float, None]) -> jnp.ndarray:
        raise NotImplementedError

    def get_initial_precision_sqrt(self,
                                   t: Union[float, None]) -> jnp.ndarray:
        raise NotImplementedError

    def get_initial_precision_det(self,
                                  t: Union[float, None]) -> float:
        raise NotImplementedError

    def initial_potential(self,
                          x: jnp.ndarray,
                          t: Union[float, None]) -> Union[float, jnp.ndarray]:
        init_mean = self.get_initial_mean(t)
        init_prec_sqrt = self.get_initial_precision_sqrt(t)
        init_prec_det = self.get_initial_precision_det(t)
        return gaussian_potential(x, init_mean, sqrt_prec=init_prec_sqrt, det_prec=init_prec_det)

    def initial_sample(self,
                       t: Union[float, None],
                       random_key: jnp.ndarray) -> Union[float, jnp.ndarray]:
        init_mean = self.get_initial_mean(t)
        init_cov_sqrt = self.get_initial_covariance_sqrt(t)
        return init_cov_sqrt @ random.normal(random_key, shape=(self.dim,)) + init_mean

    def get_transition_matrix(self,
                              t_previous: float,
                              t_new: float) -> jnp.ndarray:
        raise NotImplementedError

    def get_transition_covariance_sqrt(self,
                                       t_previous: float,
                                       t_new: float) -> jnp.ndarray:
        raise NotImplementedError

    def get_transition_precision_sqrt(self,
                                      t_previous: float,
                                      t_new: float) -> jnp.ndarray:
        raise NotImplementedError

    def get_transition_precision_det(self,
                                     t_previous: float,
                                     t_new: float) -> float:
        raise NotImplementedError

    def transition_potential(self,
                             x_previous: jnp.ndarray,
                             t_previous: float,
                             x_new: jnp.ndarray,
                             t_new: float) -> Union[float, jnp.ndarray]:
        transition_mat = self.get_transition_matrix(t_previous, t_new)
        transition_prec_sqrt = self.get_transition_precision_sqrt(t_previous, t_new)
        transition_prec_det = self.get_transition_precision_det(t_previous, t_new)
        return gaussian_potential(x_new,
                                  transition_mat @ x_previous,
                                  sqrt_prec=transition_prec_sqrt,
                                  det_prec=transition_prec_det)

    def transition_sample(self,
                          x_previous: jnp.ndarray,
                          t_previous: float,
                          t_new: float,
                          random_key: jnp.ndarray) -> jnp.ndarray:
        transition_mat = self.get_transition_matrix(t_previous, t_new)
        transition_cov_sqrt = self.get_transition_covariance_sqrt(t_previous, t_new)
        return (transition_mat @ x_previous
                + transition_cov_sqrt @ random.normal(random_key, shape=x_previous.shape)).reshape(x_previous.shape)

    def get_likelihood_matrix(self,
                              t: float) -> jnp.ndarray:
        raise NotImplementedError

    def get_likelihood_covariance_sqrt(self,
                                       t: float) -> jnp.ndarray:
        raise NotImplementedError

    def get_likelihood_precision_sqrt(self,
                                      t: float) -> jnp.ndarray:
        raise NotImplementedError

    def get_likelihood_precision_det(self,
                                     t: float) -> float:
        raise NotImplementedError

    def transition_function(self,
                            x_previous: jnp.ndarray,
                            t_previous: float,
                            t_new: float) -> jnp.ndarray:
        return self.get_transition_matrix(t_previous, t_new) @ x_previous

    def likelihood_potential(self,
                             x: jnp.ndarray,
                             y: jnp.ndarray,
                             t: float) -> Union[float, jnp.ndarray]:
        likelihood_mat = self.get_likelihood_matrix(t)
        likelihood_prec_sqrt = self.get_likelihood_precision_sqrt(t)
        likelihood_prec_det = self.get_likelihood_precision_det(t)
        return gaussian_potential(y,
                                  likelihood_mat @ x,
                                  sqrt_prec=likelihood_prec_sqrt,
                                  det_prec=likelihood_prec_det)

    def likelihood_sample(self,
                          x: jnp.ndarray,
                          t: float,
                          random_key: jnp.ndarray) -> jnp.ndarray:
        likelihood_mat = self.get_likelihood_matrix(t)
        likelihood_cov_sqrt = self.get_likelihood_covariance_sqrt(t)

        rand_shape = list(x.shape)
        rand_shape[-1] = likelihood_cov_sqrt.shape[0]
        return likelihood_mat @ x + likelihood_cov_sqrt @ random.normal(random_key, shape=rand_shape)


class TimeHomogenousLinearGaussian(LinearGaussian):
    name = 'Time-homogenous Linear Gaussian'

    # x_{t_new} ~ N(F @ x_{t_previous}, Q)
    # y_t ~ N(H @ x_t, R)
    # where F, Q, H, R are all time-homogenous
    # typically this means t is of the form jnp.arange(start, end, step)

    initial_covariance_sqrt: jnp.ndarray = None
    initial_precision_sqrt: jnp.ndarray = None
    initial_precision_det: float = None
    transition_covariance_sqrt: jnp.ndarray = None
    transition_precision_sqrt: jnp.ndarray = None
    transition_precision_det: float = None
    likelihood_covariance_sqrt: jnp.ndarray = None
    likelihood_precision_sqrt: jnp.ndarray = None
    likelihood_precision_det: float = None

    def __init__(self,
                 initial_mean: jnp.ndarray = None,
                 initial_covariance: jnp.ndarray = None,
                 transition_matrix: jnp.ndarray = None,
                 transition_covariance: jnp.ndarray = None,
                 likelihood_matrix: jnp.ndarray = None,
                 likelihood_covariance: jnp.ndarray = None,
                 name: str = None,
                 **kwargs):
        super().__init__(name=name, **kwargs)

        if 'dim' not in kwargs:
            self.dim = extract_dimension(initial_mean, initial_covariance,
                                         transition_covariance,
                                         likelihood_matrix, likelihood_covariance)
        if self.dim is None:
            raise AttributeError(f'Could not find dimension for {self.__class__.__name__}')

        if initial_mean is None:
            initial_mean = jnp.zeros(self.dim)
        self.initial_mean = initial_mean

        if initial_covariance is None:
            initial_covariance = jnp.eye(self.dim)
        self.initial_covariance = initial_covariance

        if transition_matrix is None:
            transition_matrix = jnp.eye(self.dim)
        self.transition_matrix = transition_matrix

        if transition_covariance is None:
            transition_covariance = jnp.eye(self.dim)
        self.transition_covariance = transition_covariance

        if likelihood_matrix is None:
            likelihood_matrix = jnp.eye(self.dim)
        self.likelihood_matrix = likelihood_matrix

        self.dim_obs = self.likelihood_matrix.shape[0]

        if likelihood_covariance is None:
            likelihood_covariance = jnp.eye(self.dim_obs)
        self.likelihood_covariance = likelihood_covariance

    def __setattr__(self,
                    key: str,
                    value: Any):
        self.__dict__[key] = value
        if key[-10:] == 'covariance':
            reset_covariance(self, key, value)

    def get_initial_mean(self,
                         t: Union[float, None]) -> jnp.ndarray:
        return self.initial_mean

    def get_initial_covariance_sqrt(self,
                                    t: Union[float, None]) -> jnp.ndarray:
        return self.initial_covariance_sqrt

    def get_initial_precision_sqrt(self,
                                   t: Union[float, None]) -> jnp.ndarray:
        return self.initial_precision_sqrt

    def get_initial_precision_det(self,
                                  t: Union[float, None]) -> float:
        return self.initial_precision_det

    def get_transition_matrix(self,
                              t_previous: float,
                              t_new: float) -> jnp.ndarray:
        return self.transition_matrix

    def get_transition_covariance_sqrt(self,
                                       t_previous: float,
                                       t_new: float) -> jnp.ndarray:
        return self.transition_covariance_sqrt

    def get_transition_precision_sqrt(self,
                                      t_previous: float,
                                      t_new: float) -> jnp.ndarray:
        return self.transition_precision_sqrt

    def get_transition_precision_det(self,
                                     t_previous: float,
                                     t_new: float) -> float:
        return self.transition_precision_det

    def get_likelihood_matrix(self,
                              t: float) -> jnp.ndarray:
        return self.likelihood_matrix

    def get_likelihood_covariance_sqrt(self,
                                       t: float) -> jnp.ndarray:
        return self.likelihood_covariance_sqrt

    def get_likelihood_precision_sqrt(self,
                                      t: float) -> jnp.ndarray:
        return self.likelihood_precision_sqrt

    def get_likelihood_precision_det(self,
                                     t: float) -> float:
        return self.likelihood_precision_det



########################################################################################################################
# Module: ssm/filtering.py
# Description: Particle filtering inc bootstrap filter.
#
# Web: https://github.com/SamDuffield/mocat
########################################################################################################################

from functools import partial
from typing import Union, Tuple

import jax.numpy as jnp
from jax import random, vmap, jit
from jax.lax import cond, scan

from mocat.src.core import cdict
from mocat.src.ssm.ssm import StateSpaceModel
from mocat.src.metrics import ess_log_weight


class ParticleFilter:
    name = 'Particle Filter'

    def __init__(self,
                 name: str = None):
        if name is not None:
            self.name = name

    def __repr__(self):
        return f"mocat.ParticleFilter.{self.__class__.__name__}"

    def startup(self,
                ssm_scenario: StateSpaceModel):
        pass

    def initial_potential(self,
                          ssm_scenario: StateSpaceModel,
                          x: jnp.ndarray,
                          y: jnp.ndarray,
                          t: float) -> Union[float, jnp.ndarray]:
        return ssm_scenario.initial_potential(x, t)

    def initial_sample(self,
                       ssm_scenario: StateSpaceModel,
                       y: jnp.ndarray,
                       t: float,
                       random_key: jnp.ndarray) -> jnp.ndarray:
        return ssm_scenario.initial_sample(t, random_key)

    def initial_log_weight(self,
                           ssm_scenario: StateSpaceModel,
                           x: jnp.ndarray,
                           y: jnp.ndarray,
                           t: float) -> Union[float, jnp.ndarray]:
        return -ssm_scenario.likelihood_potential(x, y, t)

    @partial(jit, static_argnums=(0, 1))
    def initial_sample_vectorised(self,
                                  ssm_scenario: StateSpaceModel,
                                  y: jnp.ndarray,
                                  t: float,
                                  random_keys: jnp.ndarray) -> jnp.ndarray:
        init_vals = vmap(self.initial_sample, (None, None, None, 0))(ssm_scenario, y, t, random_keys)
        return init_vals

    @partial(jit, static_argnums=(0, 1))
    def initial_sample_and_weight_vectorised(self,
                                             ssm_scenario: StateSpaceModel,
                                             y: jnp.ndarray,
                                             t: float,
                                             random_keys: jnp.ndarray) -> Tuple[jnp.ndarray, jnp.ndarray]:
        init_vals = vmap(self.initial_sample, (None, None, None, 0))(ssm_scenario, y, t, random_keys)

        if y is not None:
            init_log_weight = vmap(self.initial_log_weight, (None, 0, None, None))(ssm_scenario, init_vals, y, t)
        else:
            init_log_weight = jnp.zeros(len(init_vals))

        return init_vals, init_log_weight

    def proposal_potential(self,
                           ssm_scenario: StateSpaceModel,
                           x_previous: jnp.ndarray,
                           t_previous: float,
                           x_new: jnp.ndarray,
                           y_new: jnp.ndarray,
                           t_new: float) -> Union[float, jnp.ndarray]:
        raise AttributeError(f'{self.__class__.__name__} transition_potential not implemented')

    def proposal_sample(self,
                        ssm_scenario: StateSpaceModel,
                        x_previous: jnp.ndarray,
                        t_previous: float,
                        y_new: jnp.ndarray,
                        t_new: float,
                        random_key: jnp.ndarray) -> jnp.ndarray:
        raise AttributeError(f'{self.__class__.__name__} proposal_sample not implemented')

    def intermediate_log_weight(self,
                                ssm_scenario: StateSpaceModel,
                                x_previous: jnp.ndarray,
                                t_previous: float,
                                x_new: jnp.ndarray,
                                y_new: jnp.ndarray,
                                t_new: float) -> Union[float, jnp.ndarray]:
        return - ssm_scenario.transition_potential(x_previous, t_previous,
                                                   x_new, t_new) \
               - ssm_scenario.likelihood_potential(x_new, y_new, t_new) \
               + self.proposal_potential(ssm_scenario, x_previous, t_previous, x_new, y_new, t_new)

    @partial(jit, static_argnums=(0, 1))
    def proposal_sample_vectorised(self,
                                   ssm_scenario: StateSpaceModel,
                                   x_previous: jnp.ndarray,
                                   t_previous: float,
                                   y_new: jnp.ndarray,
                                   t_new: float,
                                   random_keys: jnp.ndarray) -> jnp.ndarray:
        x_new = vmap(self.proposal_sample, (None, 0, None, None, None, 0))(ssm_scenario,
                                                                           x_previous, t_previous,
                                                                           y_new, t_new,
                                                                           random_keys)
        return x_new

    @partial(jit, static_argnums=(0, 1))
    def propose_and_intermediate_weight_vectorised(self,
                                                   ssm_scenario: StateSpaceModel,
                                                   x_previous: jnp.ndarray,
                                                   t_previous: float,
                                                   y_new: jnp.ndarray,
                                                   t_new: float,
                                                   random_keys: jnp.ndarray) -> Tuple[jnp.ndarray, jnp.ndarray]:
        x_new = vmap(self.proposal_sample, (None, 0, None, None, None, 0))(ssm_scenario,
                                                                           x_previous, t_previous,
                                                                           y_new, t_new,
                                                                           random_keys)
        log_weight_new = vmap(self.intermediate_log_weight, (None, 0, None, 0, None, None))(ssm_scenario,
                                                                                            x_previous, t_previous,
                                                                                            x_new, y_new, t_new)
        return x_new, log_weight_new


class BootstrapFilter(ParticleFilter):
    name = 'Bootstrap Filter'

    def proposal_potential(self,
                           ssm_scenario: StateSpaceModel,
                           x_previous: jnp.ndarray,
                           t_previous: float,
                           x_new: jnp.ndarray,
                           y_new: jnp.ndarray,
                           t_new: float) -> Union[float, jnp.ndarray]:
        return ssm_scenario.transition_potential(x_previous, t_previous, x_new, t_new)

    def proposal_sample(self,
                        ssm_scenario: StateSpaceModel,
                        x_previous: jnp.ndarray,
                        t_previous: float,
                        y_new: jnp.ndarray,
                        t_new: float,
                        random_key: jnp.ndarray) -> jnp.ndarray:
        return ssm_scenario.transition_sample(x_previous, t_previous, t_new, random_key)

    def intermediate_log_weight(self,
                                ssm_scenario: StateSpaceModel,
                                x_previous: jnp.ndarray,
                                t_previous: float,
                                x_new: jnp.ndarray,
                                y_new: jnp.ndarray,
                                t_new: float) -> Union[float, jnp.ndarray]:
        return -ssm_scenario.likelihood_potential(x_new, y_new, t_new)


def initiate_particles(ssm_scenario: StateSpaceModel,
                       particle_filter: ParticleFilter,
                       n: int,
                       random_key: jnp.ndarray,
                       y: jnp.ndarray = None,
                       t: float = None) -> cdict:
    particle_filter.startup(ssm_scenario)

    sub_keys = random.split(random_key, n)

    init_vals, init_log_weight = particle_filter.initial_sample_and_weight_vectorised(ssm_scenario, y, t, sub_keys)

    if init_vals.ndim == 1:
        init_vals = init_vals[..., jnp.newaxis]

    initial_sample = cdict(value=init_vals[jnp.newaxis],
                           log_weight=init_log_weight[jnp.newaxis],
                           t=jnp.atleast_1d(t) if t is not None else jnp.zeros(1),
                           y=y[jnp.newaxis] if y is not None else None,
                           ess=jnp.atleast_1d(ess_log_weight(init_log_weight)))
    return initial_sample


def _resample(x_w_r: Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray]) -> jnp.ndarray:
    x, log_weight, random_key = x_w_r
    n = x.shape[-2]
    return x[..., random.categorical(random_key, log_weight, shape=(n,)), :]


def resample_particles(particles: cdict,
                       random_key: jnp.ndarray,
                       resample_full: bool = True) -> cdict:
    out_particles = particles.copy()
    n = particles.value.shape[-2]
    if out_particles.log_weight.ndim == 1:
        out_particles.value = _resample((particles.value, particles.log_weight, random_key))
        out_particles.log_weight = jnp.zeros(n)
    elif resample_full:
        out_particles.value = _resample((particles.value, particles.log_weight[-1], random_key))
        out_particles.log_weight = out_particles.log_weight.at[-1].set(jnp.zeros(n))
    else:
        latest_value = _resample((particles.value[-1], particles.log_weight[-1], random_key))
        out_particles.value = out_particles.value.at[-1].set(latest_value)
        out_particles.log_weight = out_particles.log_weight.at[-1].set(jnp.zeros(n))
    return out_particles


def propagate_particle_filter(ssm_scenario: StateSpaceModel,
                              particle_filter: ParticleFilter,
                              particles: cdict,
                              y_new: jnp.ndarray,
                              t_new: float,
                              random_key: jnp.ndarray,
                              ess_threshold: float = 0.5,
                              resample_full: bool = True) -> cdict:
    n = particles.value.shape[1]
    ess_previous = particles.ess[-1]
    out_particles = cond(ess_previous < ess_threshold * n,
                         lambda p: resample_particles(p, random_key, resample_full),
                         lambda p: p,
                         particles)

    x_previous = out_particles.value[-1]
    log_weight_previous = out_particles.log_weight[-1]
    t_previous = out_particles.t[-1]

    split_keys = random.split(random_key, n)

    x_new, log_weight_new = particle_filter.propose_and_intermediate_weight_vectorised(ssm_scenario,
                                                                                       x_previous, t_previous,
                                                                                       y_new, t_new, split_keys)

    log_weight_new = log_weight_previous + log_weight_new

    out_particles.value = jnp.append(out_particles.value, x_new[jnp.newaxis], axis=0)
    out_particles.log_weight = jnp.append(out_particles.log_weight, log_weight_new[jnp.newaxis], axis=0)
    out_particles.y = jnp.append(out_particles.y, y_new[jnp.newaxis])
    out_particles.t = jnp.append(out_particles.t, t_new)
    out_particles.ess = jnp.append(out_particles.ess, ess_log_weight(log_weight_new))
    return out_particles


def run_particle_filter_for_marginals(ssm_scenario: StateSpaceModel,
                                      particle_filter: ParticleFilter,
                                      y: jnp.ndarray,
                                      t: jnp.ndarray,
                                      random_key: jnp.ndarray,
                                      n: int = None,
                                      initial_sample: cdict = None,
                                      ess_threshold: float = 0.5) -> cdict:
    if y.ndim == 1:
        y = y[..., jnp.newaxis]

    if initial_sample is None:
        random_key, sub_key = random.split(random_key)
        init_y = y[0]
        y = y[1:]

        initial_sample = initiate_particles(ssm_scenario, particle_filter, n, sub_key, init_y, t[0])
        t = t[1:]

    if n is None:
        n = initial_sample.value.shape[1]

    num_propagate_steps = len(y)
    int_rand_keys = random.split(random_key, num_propagate_steps)

    def particle_filter_body(samps_previous: cdict,
                             iter_ind: int) -> Tuple[cdict, cdict]:
        x_previous = samps_previous.value
        log_weight_previous = samps_previous.log_weight
        int_rand_key = int_rand_keys[iter_ind]

        ess_previous = samps_previous.ess
        resample_bool = ess_previous < (ess_threshold * n)
        x_res = cond(resample_bool,
                     _resample,
                     lambda tup: tup[0],
                     (x_previous, log_weight_previous, int_rand_key))
        log_weight_res = jnp.where(resample_bool, jnp.zeros(n), log_weight_previous)

        split_keys = random.split(int_rand_key, len(x_previous))

        x_new, log_weight_new = particle_filter.propose_and_intermediate_weight_vectorised(ssm_scenario,
                                                                                           x_res,
                                                                                           samps_previous.t,
                                                                                           y[iter_ind],
                                                                                           t[iter_ind],
                                                                                           split_keys)

        log_weight_new = log_weight_res + log_weight_new

        samps_new = samps_previous.copy()
        samps_new.value = x_new
        samps_new.log_weight = log_weight_new
        samps_new.y = y[iter_ind]
        samps_new.t = t[iter_ind]
        samps_new.ess = ess_log_weight(log_weight_new)
        return samps_new, samps_new

    _, after_init_samps = scan(particle_filter_body,
                               initial_sample[-1],
                               jnp.arange(num_propagate_steps))

    out_samps = initial_sample.copy()
    out_samps.value = jnp.append(initial_sample.value, after_init_samps.value, axis=0)
    out_samps.log_weight = jnp.append(initial_sample.log_weight, after_init_samps.log_weight, axis=0)
    out_samps.y = jnp.append(initial_sample.y, after_init_samps.y, axis=0)
    out_samps.t = jnp.append(initial_sample.t, after_init_samps.t)
    out_samps.ess = jnp.append(initial_sample.ess, after_init_samps.ess)

    return out_samps



########################################################################################################################
# Module: core.py
# Description: mocat primitives Scenario and cdict.
#
# Web: https://github.com/SamDuffield/mocat
########################################################################################################################

import pickle
import copy
from typing import Union, Callable, Tuple
from pathlib import Path
from warnings import warn

import jax.numpy as jnp
from jax import grad, value_and_grad
from jax.tree_util import register_pytree_node_class
from jax.util import unzip2


@register_pytree_node_class
class cdict:
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)

    def copy(self) -> 'cdict':
        return cdict(**self.__dict__)

    def deepcopy(self) -> 'cdict':
        return copy.deepcopy(self)

    def __repr__(self):
        return f"mocat.cdict({self.__dict__.__repr__()})"

    def save(self,
             path: Union[str, Path],
             overwrite: bool = False):
        save_cdict(self, path, overwrite)

    def tree_flatten(self):
        return unzip2(self.__dict__.items())[::-1]

    @classmethod
    def tree_unflatten(cls, keys, xs):
        return cls(**dict(zip(keys, xs)))

    def __getitem__(self,
                    item: Union[str, int, slice, jnp.ndarray]) -> 'cdict':
        if isinstance(item, str):
            return self.__dict__[item]

        out_cdict = self.copy()
        for key, attr in out_cdict.__dict__.items():
            if (isinstance(attr, jnp.ndarray) and attr.ndim > 0) \
                    or (isinstance(attr, cdict) and not isinstance(attr, static_cdict)):
                out_cdict.__setattr__(key, attr[item])
        return out_cdict

    def __add__(self,
                other: 'cdict') -> 'cdict':
        out_cdict = self.copy()
        if other is None:
            return out_cdict
        for key, attr in out_cdict.__dict__.items():
            if hasattr(other, key):
                attr_atl = attr
                other_attr_atl = other.__dict__[key]
                if (isinstance(attr, jnp.ndarray) or isinstance(getattr(other, key), jnp.ndarray)):
                    out_cdict.__setattr__(key, jnp.append(jnp.atleast_1d(attr_atl),
                                                          jnp.atleast_1d(other_attr_atl), axis=0))
                elif ((isinstance(attr, cdict) and not isinstance(attr, static_cdict)) \
                      and isinstance(other_attr_atl, cdict) and not isinstance(other_attr_atl, static_cdict)) \
                        or key == 'time':
                    out_cdict.__setattr__(key, attr_atl + other_attr_atl)
        return out_cdict

    @property
    def is_empty(self):
        return self.__dict__ == {}

    def keys(self):
        return self.__dict__.keys()

    def __iter__(self):
        return self.__dict__.__iter__()


class static_cdict(cdict):
    pass


def save_cdict(in_cdict: cdict,
               path: Union[str, Path],
               overwrite: bool = False):
    path = Path(path)
    if path.suffix != '.cdict':
        path = path.with_suffix('.cdict')
    path.parent.mkdir(parents=True, exist_ok=True)
    if path.exists():
        if overwrite:
            path.unlink()
        else:
            raise RuntimeError(f'File {path} already exists.')
    with open(path, 'wb') as file:
        pickle.dump(in_cdict, file)


def load_cdict(path: Union[str, Path]) -> cdict:
    path = Path(path)
    if not path.is_file():
        raise ValueError(f'Not a file: {path}')
    if path.suffix != '.cdict':
        raise ValueError(f'Not a .cdict file: {path}')
    with open(path, 'rb') as file:
        data = pickle.load(file)

    # Convert to DeviceArray (default loads as jnp.ndarray, although I expect JAX to update this at some point)
    for attr_key, attr in data.__dict__.items():
        if isinstance(attr, jnp.ndarray):
            data.__dict__[attr_key] = jnp.asarray(attr)

    return data


def impl_checkable(func: Callable) -> Callable:
    func.not_implemented = True
    return func


def is_implemented(func):
    return not hasattr(func, 'not_implemented')


def clean_1d(scen, attr):
    if hasattr(scen, attr) and is_implemented(getattr(scen, attr)):
        setattr(scen, '_' + attr, getattr(scen, attr))

        def cleaned_func(x: Union[jnp.ndarray, float],
                         random_key: jnp.ndarray) -> float:
            x = jnp.atleast_1d(x)
            x = jnp.reshape(x, (x.size, 1))
            return jnp.squeeze(getattr(scen, '_' + attr)(x, random_key))

        setattr(scen, attr, cleaned_func)


class Scenario:
    name: str
    dim: int
    temperature: float

    grad_potential: Callable[[jnp.ndarray, jnp.ndarray], jnp.ndarray]
    potential_and_grad: Callable[[jnp.ndarray, jnp.ndarray], Tuple[float, jnp.ndarray]]
    grad_prior_potential: Callable[[jnp.ndarray, jnp.ndarray], jnp.ndarray]
    prior_potential_and_grad: Callable[[jnp.ndarray, jnp.ndarray], Tuple[float, jnp.ndarray]]
    grad_likelihood_potential: Callable[[jnp.ndarray, jnp.ndarray], jnp.ndarray]
    likelihood_potential_and_grad: Callable[[jnp.ndarray, jnp.ndarray], Tuple[float, jnp.ndarray]]

    grad_tempered_potential: Callable[[jnp.ndarray, float, jnp.ndarray], jnp.ndarray]
    tempered_potential_and_grad: Callable[[jnp.ndarray, float, jnp.ndarray], Tuple[float, jnp.ndarray]]

    def __init__(self,
                 name: str = None,
                 init_grad: bool = True,
                 **kwargs):
        if name is not None:
            self.name = name

        for key, value in kwargs.items():
            if hasattr(self, key):
                self.__dict__[key] = value

        if is_implemented(self.potential) or is_implemented(self.likelihood_potential):
            if self.dim == 1:
                clean_1d(self, 'potential')
                clean_1d(self, 'prior_potential')
                clean_1d(self, 'likelihood_potential')

            if not is_implemented(self.likelihood_potential) and is_implemented(self.potential):
                if not is_implemented(self.prior_potential):
                    self.prior_potential = lambda x, rk=None: jnp.array(0.)
                self.likelihood_potential = self.potential

            if not hasattr(self, 'temperature'):
                self.temperature = 1.

            if not is_implemented(self.prior_potential):
                warn(f'{self.name} prior_potential not initiated, assuming uniform')
                self.prior_potential = lambda x, random_key=None: jnp.array(0.)

            self.tempered_potential \
                = lambda x, temperature, random_key=None: self.prior_potential(x, random_key) \
                                                          + temperature * self.likelihood_potential(x, random_key)

            self.potential = lambda x, random_key=None: self.tempered_potential(x, self.temperature, random_key)

        if init_grad:
            self.init_grad()

    def __setattr__(self, key, value):
        self.__dict__[key] = value
        if key in ('potential', 'prior_potential', 'likelihood_potential') and hasattr(self, 'grad_potential'):
            self.init_grad()

    def init_grad(self):
        self.grad_prior_potential = grad(self.prior_potential)
        self.prior_potential_and_grad = value_and_grad(self.prior_potential)

        self.grad_likelihood_potential = grad(self.likelihood_potential)
        self.likelihood_potential_and_grad = value_and_grad(self.likelihood_potential)

        self.grad_tempered_potential \
            = lambda x, temperature, random_key=None: self.grad_prior_potential(x, random_key) \
                                                      + temperature * self.grad_likelihood_potential(x, random_key)

        def tempered_potential_and_grad(x: jnp.ndarray,
                                        temperature: float,
                                        random_key: jnp.ndarray) -> Tuple[float, jnp.ndarray]:
            prior_pot, prior_grad = self.prior_potential_and_grad(x, random_key)
            lik_pot, lik_grad = self.likelihood_potential_and_grad(x, random_key)
            return prior_pot + temperature * lik_pot, \
                   prior_grad + temperature * lik_grad

        self.tempered_potential_and_grad = tempered_potential_and_grad

        self.grad_potential = lambda x, random_key=None: self.grad_tempered_potential(x,
                                                                                      self.temperature,
                                                                                      random_key)
        self.potential_and_grad = lambda x, random_key=None: self.tempered_potential_and_grad(x,
                                                                                              self.temperature,
                                                                                              random_key)

    def __repr__(self):
        return f"mocat.Scenario.{self.__class__.__name__}({self.__dict__.__repr__()})"

    @impl_checkable
    def potential(self, x: jnp.ndarray,
                  random_key: jnp.ndarray = None) -> float:
        raise AttributeError(f'{self.name} potential not initiated')

    @impl_checkable
    def prior_potential(self,
                        x: jnp.ndarray,
                        random_key: jnp.ndarray = None) -> float:
        raise AttributeError(f'{self.name} prior_potential not initiated')

    @impl_checkable
    def prior_sample(self,
                     random_key: jnp.ndarray) -> Union[float, jnp.ndarray]:
        raise AttributeError(f'{self.name} prior_sample not initiated')

    @impl_checkable
    def likelihood_potential(self,
                             x: jnp.ndarray,
                             random_key: jnp.ndarray = None) -> float:
        raise AttributeError(f'{self.name} prior_potential not initiated')

    @impl_checkable
    def likelihood_sample(self,
                          x: jnp.ndarray,
                          random_key: jnp.ndarray) -> jnp.ndarray:
        raise AttributeError(f'{self.name} likelihood_sample not initiated')



########################################################################################################################
# Module: tests/test_toy_scenarios.py
# Description: Tests for bivariate toy scenarios
#
# Web: https://github.com/SamDuffield/mocat
########################################################################################################################


import unittest

import jax.numpy as jnp
import numpy.testing as npt

from mocat.src.scenarios.twodim import toy_examples


class TestGaussian(unittest.TestCase):
    scenario = toy_examples.Gaussian()

    def test_basic(self):
        npt.assert_equal(self.scenario.dim, 2)
        npt.assert_array_equal(self.scenario.mean, jnp.zeros(2))
        npt.assert_array_equal(self.scenario.covariance, jnp.eye(2))

    def test_vec_potential(self):
        npt.assert_array_equal(self.scenario.vec_potential(jnp.array([0, 2, 4]),
                                                           jnp.array([1, 3, 5])),
                               jnp.array([[0.5, 2.5, 8.5],
                                          [4.5, 6.5, 12.5],
                                          [12.5, 14.5, 20.5]]))

    def test_auto_axes_lims(self):
        self.scenario.auto_axes_lims()
        npt.assert_almost_equal(-self.scenario.xlim[0], self.scenario.xlim[1], decimal=4)
        npt.assert_almost_equal(self.scenario.xlim[0], self.scenario.ylim[0], decimal=4)
        npt.assert_almost_equal(self.scenario.xlim[1], self.scenario.ylim[1], decimal=4)

        npt.assert_(self.scenario.xlim[1] > 1)
        npt.assert_(self.scenario.xlim[1] < 5)

        self.scenario.covariance = jnp.array([[10., 9.], [9., 50.]])
        self.scenario.auto_axes_lims()

        npt.assert_almost_equal(-self.scenario.xlim[0], self.scenario.xlim[1], decimal=4)
        npt.assert_almost_equal(-self.scenario.ylim[0], self.scenario.ylim[1], decimal=4)

        npt.assert_(self.scenario.xlim[1] > 5)
        npt.assert_(self.scenario.xlim[1] < 15)

        npt.assert_(self.scenario.ylim[1] > 10)
        npt.assert_(self.scenario.ylim[1] < 20)

        self.scenario.covariance = jnp.eye(2)


class TestBanana(unittest.TestCase):
    scenario = toy_examples.Banana()
    scenario.curviness = 0.03
    scenario.lengthiness = 100

    def test_basic(self):
        npt.assert_equal(self.scenario.dim, 2)

    def test_potential(self):
        npt.assert_array_equal(self.scenario.potential(jnp.zeros(2)), 4.5)
        npt.assert_array_equal(self.scenario.potential(jnp.arange(2)), 2.)

    def test_grad_potential(self):
        npt.assert_array_equal(self.scenario.grad_potential(jnp.zeros(2)), [0., -3.])
        npt.assert_array_equal(self.scenario.grad_potential(jnp.arange(2, dtype='float32')), [0., -2.])

    def test_vec_potential(self):
        npt.assert_array_almost_equal(self.scenario.vec_potential(jnp.array([0, 2, 4]),
                                                                  jnp.array([1, 3, 5])),
                                      jnp.array([[2., 1.7872, 1.2352],
                                                 [0., 0.0272, 0.1952],
                                                 [2., 2.2672, 3.1552]]))

    def test_auto_axes_lims(self):
        self.scenario.auto_axes_lims()
        npt.assert_almost_equal(-self.scenario.xlim[0], self.scenario.xlim[1], decimal=4)
        npt.assert_(abs(self.scenario.ylim[0]) > self.scenario.ylim[1])

        npt.assert_(self.scenario.xlim[1] > 15)
        npt.assert_(self.scenario.xlim[1] < 30)

        npt.assert_(self.scenario.ylim[0] < -5)
        npt.assert_(self.scenario.ylim[0] > -20)

        npt.assert_(self.scenario.ylim[1] > 1)
        npt.assert_(self.scenario.ylim[1] < 15)


class TestNealFunnel(unittest.TestCase):
    scenario = toy_examples.NealFunnel()

    def test_basic(self):
        npt.assert_equal(self.scenario.dim, 2)

    def test_potential(self):
        npt.assert_array_equal(self.scenario.potential(jnp.zeros(2)), 0.)
        npt.assert_array_almost_equal(self.scenario.potential(jnp.arange(2)), 0.166667)

    def test_grad_potential(self):
        npt.assert_array_almost_equal(self.scenario.grad_potential(jnp.zeros(2)), jnp.zeros(2))
        npt.assert_array_almost_equal(self.scenario.grad_potential(jnp.arange(2, dtype='float32')),
                                      jnp.array([0., 0.333333]))

    def test_vec_potential(self):
        npt.assert_array_almost_equal(self.scenario.vec_potential(jnp.array([0, 2, 4]),
                                                                  jnp.array([1, 3, 5])),
                                      jnp.array([[0.166667, 1.379728, 5.018912],
                                                 [1.5, 1.94626, 3.285041],
                                                 [4.166667, 4.330836, 4.823347]]))

    def test_auto_axes_lims(self):
        self.scenario.auto_axes_lims()
        npt.assert_almost_equal(-self.scenario.xlim[0], self.scenario.xlim[1], decimal=4)
        npt.assert_almost_equal(-self.scenario.ylim[0], self.scenario.ylim[1], decimal=4)

        npt.assert_(self.scenario.xlim[1] > 1)
        npt.assert_(self.scenario.xlim[1] < 10)

        npt.assert_(self.scenario.ylim[1] > 1)
        npt.assert_(self.scenario.ylim[1] < 20)


class TestGaussianMixture(unittest.TestCase):
    scenario = toy_examples.GaussianMixture()

    def test_basic(self):
        npt.assert_equal(self.scenario.dim, 2)
        npt.assert_equal(self.scenario.means.ndim, 2)
        npt.assert_equal(self.scenario.means.shape[-1], 2)
        npt.assert_equal(self.scenario.covariances.ndim, 3)
        npt.assert_equal(self.scenario.covariances.shape[-1], 2)
        npt.assert_equal(self.scenario.precisions.ndim, 3)
        npt.assert_equal(self.scenario.precisions.shape[-1], 2)
        npt.assert_equal(self.scenario.precision_sqrts.ndim, 3)
        npt.assert_equal(self.scenario.precision_sqrts.shape[-1], 2)

    def test_vec_potential(self):
        npt.assert_array_almost_equal(self.scenario.vec_potential(jnp.array([0, 2, 4]),
                                                                  jnp.array([1, 3, 5])),
                                      jnp.array([[5.012875, 3.705686, 5.706022],
                                                 [5.031018, 3.72383, 5.724165],
                                                 [9.031024, 7.723836, 9.724172]]))

    def test_auto_axes_lims(self):
        self.scenario.auto_axes_lims()
        npt.assert_almost_equal(-self.scenario.xlim[0], self.scenario.xlim[1], decimal=4)
        npt.assert_almost_equal(self.scenario.xlim[0], self.scenario.ylim[0], decimal=4)
        npt.assert_almost_equal(self.scenario.xlim[1], self.scenario.ylim[1], decimal=4)

        npt.assert_(self.scenario.xlim[1] > 1)
        npt.assert_(self.scenario.xlim[1] < 10)

        self.scenario.covariances = jnp.array([[10., 9.], [9., 30.]])
        self.test_basic()

        self.scenario.auto_axes_lims()

        npt.assert_almost_equal(-self.scenario.xlim[0], self.scenario.xlim[1], decimal=4)
        npt.assert_almost_equal(-self.scenario.ylim[0], self.scenario.ylim[1], decimal=4)

        npt.assert_(self.scenario.xlim[1] > 7)
        npt.assert_(self.scenario.xlim[1] < 15)

        npt.assert_(self.scenario.ylim[1] > 11)
        npt.assert_(self.scenario.ylim[1] < 16)

        self.scenario.covariances = jnp.eye(2)


class TestDoubleWell(unittest.TestCase):
    scenario = toy_examples.DoubleWell()

    def test_basic(self):
        npt.assert_equal(self.scenario.dim, 2)

    def test_vec_potential(self):
        npt.assert_array_almost_equal(self.scenario.vec_potential(jnp.array([0, 2, 4]),
                                                                  jnp.array([1, 3, 5])),
                                      jnp.array([[-0.25, 1.75, 55.75],
                                                 [15.75, 17.75, 71.75],
                                                 [143.75, 145.75, 199.75]]))

    def test_auto_axes_lims(self):
        self.scenario.auto_axes_lims()
        npt.assert_almost_equal(-self.scenario.xlim[0], self.scenario.xlim[1], decimal=4)
        npt.assert_almost_equal(self.scenario.xlim[0], self.scenario.ylim[0], decimal=4)
        npt.assert_almost_equal(self.scenario.xlim[1], self.scenario.ylim[1], decimal=4)

        npt.assert_(self.scenario.xlim[1] > 1)
        npt.assert_(self.scenario.xlim[1] < 5)


if __name__ == '__main__':
    unittest.main()



########################################################################################################################
# Module: ssm/tests/test_ssm_linear_gaussian.py
# Description: Tests for linear TemporalGaussian state space models (and Kalman filtering/smoothing).
#
# Web: https://github.com/SamDuffield/mocat
########################################################################################################################

import unittest

import jax.numpy as jnp
from jax import random
import numpy.testing as npt
from mocat.src.ssm.ssm import StateSpaceModel
from mocat.src.ssm.filtering import BootstrapFilter, run_particle_filter_for_marginals, initiate_particles
from mocat.src.ssm.backward import backward_simulation
from mocat.src.ssm.online_smoothing import propagate_particle_smoother


class TestSSM(unittest.TestCase):
    ssm_scenario: StateSpaceModel
    len_t: int = 20
    t: jnp.ndarray = jnp.arange(len_t, dtype='float32')
    max_rejections: int = 2
    n: int = int(2e3)

    def _test_simulate(self):

        self.sim_samps = self.ssm_scenario.simulate(self.t, random.PRNGKey(0))
        npt.assert_array_equal(self.sim_samps.t, self.t)
        self.assertEqual(self.len_t, self.sim_samps.x.shape[0])
        self.assertEqual(self.len_t, self.sim_samps.y.shape[0])

    def _test_bootstrap(self):
        if not hasattr(self, 'sim_samps'):
            self.sim_samps = self.ssm_scenario.simulate(self.t, random.PRNGKey(0))
        self.pf_samps = run_particle_filter_for_marginals(self.ssm_scenario,
                                                          BootstrapFilter(),
                                                          self.sim_samps.y,
                                                          self.t,
                                                          random.PRNGKey(0),
                                                          n=self.n)

        npt.assert_array_less(self.sim_samps.x[:, 0], jnp.max(self.pf_samps.value, axis=1)[:, 0])
        npt.assert_array_less(jnp.min(self.pf_samps.value, axis=1)[:, 0], self.sim_samps.x[:, 0])

    def backward_preprocess(self):
        if not hasattr(self, 'sim_samps'):
            self.sim_samps = self.ssm_scenario.simulate(self.t, random.PRNGKey(0))
        if not hasattr(self, 'pf_samps'):
            self.pf_samps = run_particle_filter_for_marginals(self.ssm_scenario,
                                                              BootstrapFilter(),
                                                              self.sim_samps.y,
                                                              self.t,
                                                              random.PRNGKey(0),
                                                              n=int(1e4))

    def backward_postprocess(self):
        npt.assert_array_less(self.sim_samps.x[:, 0], jnp.max(self.backward_samps.value, axis=1)[:, 0])
        npt.assert_array_less(jnp.min(self.backward_samps.value, axis=1)[:, 0], self.sim_samps.x[:, 0])
        self.assertFalse(hasattr(self.backward_samps, 'log_weight'))

    def _test_ffbsi_full(self):
        self.backward_preprocess()
        self.backward_samps = backward_simulation(self.ssm_scenario,
                                                  self.pf_samps,
                                                  random.PRNGKey(0),
                                                  self.n,
                                                  maximum_rejections=0)
        self.backward_postprocess()

    def _test_ffbsi_rejection(self):
        self.backward_preprocess()
        self.backward_samps = backward_simulation(self.ssm_scenario,
                                                  self.pf_samps,
                                                  random.PRNGKey(0),
                                                  self.n,
                                                  maximum_rejections=self.max_rejections)
        self.backward_postprocess()

    def _test_backward(self):
        self._test_ffbsi_full()
        self._test_ffbsi_rejection()

    def _test_online_smoothing_pf_full(self):
        if not hasattr(self, 'sim_samps'):
            self.sim_samps = self.ssm_scenario.simulate(self.t, random.PRNGKey(0))
        pf = BootstrapFilter()
        len_t = len(self.t)
        rkeys = random.split(random.PRNGKey(0), len_t)

        particles = initiate_particles(self.ssm_scenario, pf, self.n,
                                       rkeys[0], self.sim_samps.y[0], self.t[0])
        for i in range(1, len_t):
            particles = propagate_particle_smoother(self.ssm_scenario, pf, particles,
                                                    self.sim_samps.y[i], self.t[i], rkeys[i], 3,
                                                    False)
        npt.assert_array_less(((self.sim_samps.x[:, 0] - jnp.max(particles.value, axis=1)[:, 0]) > 0).mean(), 0.1)
        npt.assert_array_less(((self.sim_samps.x[:, 0] - jnp.min(particles.value, axis=1)[:, 0]) < 0).mean(), 0.1)

    def _test_online_smoothing_pf_rejection(self):
        if not hasattr(self, 'sim_samps'):
            self.sim_samps = self.ssm_scenario.simulate(self.t, random.PRNGKey(0))
        pf = BootstrapFilter()
        len_t = len(self.t)
        rkeys = random.split(random.PRNGKey(0), len_t)

        particles = initiate_particles(self.ssm_scenario, pf, self.n,
                                       rkeys[0], self.sim_samps.y[0], self.t[0])
        for i in range(1, len_t):
            particles = propagate_particle_smoother(self.ssm_scenario, pf, particles,
                                                    self.sim_samps.y[i], self.t[i], rkeys[i], 3,
                                                    False, maximum_rejections=10)

        npt.assert_array_less(((self.sim_samps.x[:, 0] - jnp.max(particles.value, axis=1)[:, 0]) > 0).mean(), 0.1)
        npt.assert_array_less(((self.sim_samps.x[:, 0] - jnp.min(particles.value, axis=1)[:, 0]) < 0).mean(), 0.1)

    def _test_online_smoothing_bs_full(self):
        if not hasattr(self, 'sim_samps'):
            self.sim_samps = self.ssm_scenario.simulate(self.t, random.PRNGKey(0))
        pf = BootstrapFilter()
        len_t = len(self.t)
        rkeys = random.split(random.PRNGKey(0), len_t)

        particles = initiate_particles(self.ssm_scenario, pf, self.n,
                                       rkeys[0], self.sim_samps.y[0], self.t[0])
        for i in range(1, len_t):
            particles = propagate_particle_smoother(self.ssm_scenario, pf, particles,
                                                    self.sim_samps.y[i], self.t[i], rkeys[i], 3,
                                                    True)
        npt.assert_array_less(((self.sim_samps.x[:, 0] - jnp.max(particles.value, axis=1)[:, 0]) > 0).mean(), 0.1)
        npt.assert_array_less(((self.sim_samps.x[:, 0] - jnp.min(particles.value, axis=1)[:, 0]) < 0).mean(), 0.1)

    def _test_online_smoothing_bs_rejection(self):
        if not hasattr(self, 'sim_samps'):
            self.sim_samps = self.ssm_scenario.simulate(self.t, random.PRNGKey(0))
        pf = BootstrapFilter()
        len_t = len(self.t)
        rkeys = random.split(random.PRNGKey(0), len_t)

        particles = initiate_particles(self.ssm_scenario, pf, self.n,
                                       rkeys[0], self.sim_samps.y[0], self.t[0])
        for i in range(1, len_t):
            particles = propagate_particle_smoother(self.ssm_scenario, pf, particles,
                                                    self.sim_samps.y[i], self.t[i], rkeys[i], 3,
                                                    True, maximum_rejections=10)
        npt.assert_array_less(((self.sim_samps.x[:, 0] - jnp.max(particles.value, axis=1)[:, 0]) > 0).mean(), 0.1)
        npt.assert_array_less(((self.sim_samps.x[:, 0] - jnp.min(particles.value, axis=1)[:, 0]) < 0).mean(), 0.1)


if __name__ == '__main__':
    unittest.main()



########################################################################################################################
# Module: ssm/ssm.py
# Description: state-space model class
#
# Web: https://github.com/SamDuffield/mocat
########################################################################################################################

from typing import Union

import jax.numpy as jnp
from jax import vmap, grad, value_and_grad
from jax.lax import scan
from jax import random

from mocat.src.core import cdict


class StateSpaceModel:
    name = None
    dim = None
    dim_obs = None

    def __init__(self,
                 name: str = None,
                 **kwargs):
        if name is not None:
            self.name = name

        for key, value in kwargs.items():
            setattr(self, key, value)

        if not hasattr(self, 'grad_initial_potential'):
            self.grad_initial_potential = grad(self.initial_potential)
            self.initial_potential_and_grad = value_and_grad(self.initial_potential)

        if not hasattr(self, 'grad_transition_potential'):
            self.grad_transition_potential = grad(self.transition_potential, argnums=(0, 2))
            self.transition_potential_and_grad = value_and_grad(self.transition_potential, argnums=(0, 2))

        if not hasattr(self, 'grad_likelihood_potential'):
            self.grad_likelihood_potential = grad(self.likelihood_potential)
            self.likelihood_potential_and_grad = value_and_grad(self.initial_potential)

    def __repr__(self):
        return f"mocat.StateSpaceModel.{self.__class__.__name__}"

    def initial_potential(self,
                          x: jnp.ndarray,
                          t: float) -> Union[float, jnp.ndarray]:
        raise AttributeError(f'{self.__class__.__name__} initial_potential not implemented')

    def initial_sample(self,
                       t: float,
                       random_key: jnp.ndarray) -> Union[float, jnp.ndarray]:
        raise AttributeError(f'{self.__class__.__name__} initial_sample not implemented')

    def transition_potential(self,
                             x_previous: jnp.ndarray,
                             t_previous: float,
                             x_new: jnp.ndarray,
                             t_new: float) -> Union[float, jnp.ndarray]:
        raise AttributeError(f'{self.__class__.__name__} transition_potential not implemented')

    def transition_sample(self,
                          x_previous: jnp.ndarray,
                          t_previous: float,
                          t_new: float,
                          random_key: jnp.ndarray) -> jnp.ndarray:
        raise AttributeError(f'{self.__class__.__name__} transition_sample not implemented')

    def likelihood_potential(self,
                             x: jnp.ndarray,
                             y: jnp.ndarray,
                             t: float) -> Union[float, jnp.ndarray]:
        raise NotImplementedError(f'{self.__class__.__name__} likelihood_potential not implemented')

    def likelihood_sample(self,
                          x: jnp.ndarray,
                          t: float,
                          random_key: jnp.ndarray) -> jnp.ndarray:
        raise AttributeError(f'{self.__class__.__name__} transition_sample not implemented')

    def _smoothing_potential(self,
                             x_all: jnp.ndarray,
                             y_all: jnp.ndarray,
                             t_all: jnp.ndarray) -> Union[float, jnp.ndarray]:
        # x_all: shape = (T, d)
        # t_all: shape = (T,)
        # y_all: shape = (T, d_y)
        return self.initial_potential(x_all[0], t_all[0]) \
               + vmap(self.transition_potential)(x_all[:-1], t_all[:-1], x_all[1:], t_all[1:]).sum() \
               + vmap(self.likelihood_potential)(x_all, y_all, t_all).sum()

    def smoothing_potential(self,
                            x_all: jnp.ndarray,
                            y_all: jnp.ndarray,
                            t_all: jnp.ndarray) -> Union[float, jnp.ndarray]:
        if len(t_all) == 1:
            return self.initial_potential(x_all[0], t_all[0]) \
                   + self.likelihood_potential(x_all[0], y_all[0], t_all[0])
        elif len(t_all) == 2:
            trans_pot = self.transition_potential(x_all[0], t_all[0], x_all[1], t_all[1])
            return jnp.vstack([self.initial_potential(x_all[0], t_all[0]) + trans_pot[0] \
                              + self.likelihood_potential(x_all[0], y_all[0], t_all[0]),
                              trans_pot[1] + self.likelihood_potential(x_all[1], y_all[1], t_all[1])])
        else:
            return self._smoothing_potential(x_all, y_all, t_all)

    def _grad_smoothing_potential(self,
                                  x_all: jnp.ndarray,
                                  y_all: jnp.ndarray,
                                  t_all: jnp.ndarray) -> Union[float, jnp.ndarray]:
        # it may well be quicker to do grad(smoothing_potential)
        # but currently there is a jax bug that breaks on combination of scan - grad - vmap - odeint
        initial_grads = self.grad_initial_potential(x_all[0], t_all[0])
        transition_prev_grads, transition_next_grads = vmap(self.grad_transition_potential)(x_all[:-1], t_all[:-1],
                                                                                            x_all[1:], t_all[1:])
        likelihood_grads = vmap(self.grad_likelihood_potential)(x_all, y_all, t_all)
        return jnp.vstack([initial_grads + transition_prev_grads[0] + likelihood_grads[0],
                          transition_prev_grads[1:] + transition_next_grads[:-1] + likelihood_grads[1:-1],
                          transition_next_grads[-1] + likelihood_grads[-1]])

    def grad_smoothing_potential(self,
                                 x_all: jnp.ndarray,
                                 y_all: jnp.ndarray,
                                 t_all: jnp.ndarray) -> Union[float, jnp.ndarray]:
        if len(t_all) == 1:
            return (self.grad_initial_potential(x_all[0], t_all[0])
                    + self.grad_likelihood_potential(x_all[0], y_all[0], t_all[0]))[jnp.newaxis]
        elif len(t_all) == 2:
            grad_trans_pot = self.grad_transition_potential(x_all[0], t_all[0], x_all[1], t_all[1])
            return jnp.vstack([self.grad_initial_potential(x_all[0], t_all[0]) + grad_trans_pot[0] \
                              + self.grad_likelihood_potential(x_all[0], y_all[0], t_all[0]),
                              grad_trans_pot[1] + self.grad_likelihood_potential(x_all[1], y_all[1], t_all[1])])
        else:
            return self._grad_smoothing_potential(x_all, y_all, t_all)

    def simulate(self,
                 t_all: jnp.ndarray,
                 random_key: jnp.ndarray) -> cdict:

        len_t = len(t_all)

        random_keys = random.split(random_key, 2 * len_t)
        latent_keys = random_keys[:len_t]
        obs_keys = random_keys[len_t:]

        x_init = self.initial_sample(t_all[0], latent_keys[0])

        def transition_body(x, i):
            new_x = self.transition_sample(x, t_all[i - 1], t_all[i], latent_keys[i])
            return new_x, new_x

        _, x_all_but_zero = scan(transition_body,
                                 x_init,
                                 jnp.arange(1, len_t))

        x_all = jnp.append(x_init[jnp.newaxis], x_all_but_zero, axis=0)

        y = vmap(self.likelihood_sample)(x_all, t_all, obs_keys)

        out_cdict = cdict(x=x_all, y=y, t=t_all, name=f'{self.name} simulation')
        return out_cdict



# Analyse sensitivity of lag parameter for online smoothing

import matplotlib.pyplot as plt
from jax import numpy as jnp, random, jit
from scipy.interpolate import make_interp_spline
import os
from time import time
import pickle

import mocat
from mocat import ssm

fig_dir = os.path.dirname(os.getcwd()) + '/images/linear_gaussian_1D'

# Use splines to smooth trajectories
def smooth(t, x, k=3):
    t_new = jnp.linspace(min(t), max(t), 200)
    spl = make_interp_spline(t, x, k=k)
    y_smooth = spl(t_new)
    return t_new, y_smooth


# Initiate random seed
random_keys = random.split(random.PRNGKey(3), 100)

# Number of particles to generate
n = 10000

# Transition variance
sigma2_x = 1.

# Observation variance
sigma2_y = 1.

# Lags
lags = jnp.arange(2, 11)

# Rejections
max_rejs = 0

# Initiate scenario
dim = 1
lgssm = ssm.TimeHomogenousLinearGaussian(initial_mean=jnp.zeros(dim),
                                         initial_covariance=jnp.eye(dim) * sigma2_x,
                                         transition_matrix=jnp.eye(dim),
                                         transition_covariance=jnp.eye(dim) * sigma2_x,
                                         likelihood_matrix=jnp.eye(dim),
                                         likelihood_covariance=jnp.eye(dim) * sigma2_y,
                                         name='1D LGSSM')

# Simulate underlying true hidden process and observations
len_t = 40
true_process = lgssm.simulate(jnp.arange(len_t, dtype='float32'), random_keys[0])

# Initiate particle filter
pf = ssm.OptimalNonLinearGaussianParticleFilter()

# Analytical posterior
mean_x = jnp.zeros(len_t)
mean_y = jnp.zeros(len_t)
cov_x = jnp.array([[min(i, j) + 1 for i in range(len_t)] for j in range(len_t)], dtype='float32')
prec_x = jnp.linalg.inv(cov_x)
cov_y = cov_x + jnp.eye(len_t)
prec_y = jnp.linalg.inv(cov_y)
mean_post = mean_x + cov_x @ prec_y @ (true_process.y[:, 0] - mean_y)
cov_post = cov_x - cov_x @ prec_y @ cov_x
prec_post = jnp.linalg.inv(cov_post)


def metric(samp_arr, log_weight=None):
    if isinstance(samp_arr, mocat.cdict):
        samp_arr = samp_arr.value
    if samp_arr.ndim == 3:
        samp_arr = samp_arr[..., 0].T
    if log_weight is not None:
        samp_arr = samp_arr[random.categorical(random.PRNGKey(0), log_weight, shape=(len(samp_arr),))]
    mean = samp_arr.mean(0)
    cov = jnp.cov(samp_arr.T, ddof=1)
    return 0.5 * (jnp.trace(prec_post @ cov)
                  + (mean_post - mean).T @ prec_post @ (mean_post - mean)
                  - len_t + jnp.log(jnp.linalg.det(cov_post) / jnp.linalg.det(cov)))


# Run particle filter for marginals
pf_marginals = ssm.run_particle_filter_for_marginals(lgssm, pf, true_process.y, true_process.t, random_keys[1], n,
                                                     ess_threshold=1.)

# Run particle filter for trajectories
pft_rks = random.split(random_keys[2], len_t)
pf_trajectories = ssm.initiate_particles(lgssm, pf, n, pft_rks[0], true_process.y[0], true_process.t[0])
for i in range(1, len_t):
    pf_trajectories = ssm.propagate_particle_filter(lgssm, pf, pf_trajectories, true_process.y[i], true_process.t[i],
                                                    pft_rks[i], ess_threshold=1.)

# Backward simulation
start = time()
backsim = ssm.backward_simulation(lgssm, pf_marginals, random_keys[3], n)
backsim.value.block_until_ready()
backsim.time = time() - start

# Online smoother pf fixed lag
pf_fl_rk = random_keys[5]
pf_fl_samps = []
for lag in lags:
    pf_fl_rk, _ = random.split(pf_fl_rk)
    pf_fl_rks = random.split(pf_fl_rk, len_t)
    print(f'OPS - wPF - lag = {lag}')
    propagate_fl_pf = lambda i, parts: ssm.propagate_particle_smoother(lgssm, pf, parts, true_process.y[i],
                                                                       true_process.t[i], pf_fl_rks[i], lag, False,
                                                                       maximum_rejections=max_rejs)
    propagate_fl_pf = jit(propagate_fl_pf)
    start = time()
    pf_fl_trajectories = ssm.initiate_particles(lgssm, pf, n, pf_fl_rks[0], true_process.y[0], true_process.t[0])
    for i in range(1, len_t):
        pf_fl_trajectories = propagate_fl_pf(i, pf_fl_trajectories)
    pf_fl_trajectories.value.block_until_ready()
    pf_fl_trajectories.time = time() - start
    print(pf_fl_trajectories.time)
    pf_fl_trajectories.lag = lag
    pf_fl_trajectories.maximum_rejections = max_rejs
    pf_fl_samps.append(pf_fl_trajectories.copy())

with open(fig_dir + '/fl_pf_varylag.pkl', 'wb') as f:
    pickle.dump(pf_fl_samps, f)

#
# with open(fig_dir + '/fl_pf_varylag.pkl', 'rb') as f:
#     pf_fl_samps = pickle.load(f)


# Online smoother bs fixed lag
bs_fl_rk = random_keys[6]
bs_fl_samps = []
for lag in lags:
    bs_fl_rk, _ = random.split(bs_fl_rk)
    bs_fl_rks = random.split(bs_fl_rk, len_t)
    print(f'OPS - wBS - lag = {lag}')
    propagate_fl_bs = lambda i, parts: ssm.propagate_particle_smoother(lgssm, pf, parts, true_process.y[i],
                                                                       true_process.t[i], bs_fl_rks[i], lag, True,
                                                                       maximum_rejections=max_rejs, ess_threshold=1.)
    propagate_fl_bs = jit(propagate_fl_bs)
    start = time()
    bs_fl_trajectories = ssm.initiate_particles(lgssm, pf, n, bs_fl_rks[0], true_process.y[0], true_process.t[0])
    for i in range(1, len_t):
        bs_fl_trajectories = propagate_fl_bs(i, bs_fl_trajectories)
    bs_fl_trajectories.value.block_until_ready()
    bs_fl_trajectories.time = time() - start
    print(bs_fl_trajectories.time)
    bs_fl_trajectories.lag = lag
    bs_fl_trajectories.maximum_rejections = max_rejs
    bs_fl_samps.append(bs_fl_trajectories.copy())

with open(fig_dir + '/fl_bs_varylag.pkl', 'wb') as f:
    pickle.dump(bs_fl_samps, f)

# with open(fig_dir + '/fl_bs_varylag.pkl', 'rb') as f:
#     bs_fl_samps = pickle.load(f)

# Plot metrics
pf_joint_metric = metric(pf_trajectories, log_weight=pf_trajectories.log_weight[-1])
backsim_metric = metric(backsim)
pf_fl_metrics = jnp.array([metric(pf_fl_samps[j]) for j in range(len(lags))])
bs_fl_metrics = jnp.array([metric(bs_fl_samps[j]) for j in range(len(lags))])
fig, ax = plt.subplots()
ax.axhline(pf_joint_metric, color='purple', label='Particle Filter for smoothing')
ax.axhline(backsim_metric, color='forestgreen', label='FFBSi')
ax.plot(*smooth(lags, pf_fl_metrics, 2), color='red', label='Online Particle Smoother')
ax.plot(*smooth(lags, bs_fl_metrics, 2), color='blue', label='...with Backward Simulation')
ax.set_xlabel(r'$L$')
fig.legend(loc='upper left', bbox_to_anchor=(0.2, 0.9))
fig.savefig(fig_dir + f'/varylag', dpi=300)



########################################################################################################################
# Module: scenarios/toy_examples.py
# Description: Some simple multivariate scenarios
#
# Web: https://github.com/SamDuffield/mocat
########################################################################################################################

from typing import Union, Any, Callable

import jax.numpy as jnp
from jax import vmap, jit

from mocat.src.core import Scenario
from mocat.utils import reset_covariance, gaussian_potential


class Gaussian(Scenario):
    name = "Gaussian"

    covariance_sqrt: jnp.ndarray
    precision_sqrt: jnp.ndarray
    precision_det: float
    precision_mul: Callable[[jnp.ndarray, jnp.ndarray], jnp.ndarray]

    def __init__(self,
                 dim: int = 1,
                 mean: Union[float, jnp.ndarray] = None,
                 covariance: Union[float, jnp.ndarray] = None,
                 **kwargs):
        if mean is not None:
            self.dim = mean.shape[-1]
        elif covariance is not None:
            self.dim = covariance.shape[-1]
        else:
            self.dim = dim
        self.mean = jnp.zeros(dim) if mean is None else mean
        self.covariance = jnp.ones(dim) if covariance is None else covariance
        super().__init__(**kwargs)

    def likelihood_potential(self,
                             x: jnp.ndarray,
                             random_key: jnp.ndarray = None) -> Union[float, jnp.ndarray]:
        x_diff = self.precision_mul(x - self.mean, self.precision_sqrt.T)
        return 0.5 * jnp.square(x_diff).sum(axis=-1)

    def __setattr__(self,
                    key: str,
                    value: Any):
        self.__dict__[key] = value
        if key[-10:] == 'covariance':
            reset_covariance(self, key, value)


class GaussianMixture(Scenario):
    name = "Gaussian Mixture"

    def __init__(self,
                 means: jnp.ndarray,
                 covariances: jnp.ndarray = None,
                 weights: jnp.ndarray = None,
                 **kwargs):
        self.means = means[:, jnp.newaxis] if means.ndim == 1 else means
        self.dim = self.means.shape[-1]
        if weights is None:
            # Default: equal mixture weights
            self.weights = jnp.ones(self.n_components) / self.n_components
        else:
            self.weights = weights
        self.covariances = covariances

        super().__init__(**kwargs)

    def __setattr__(self, key, value):
        self.__dict__[key] = value
        if key == 'covariances':
            if value is None:
                # Default: identity covariances
                self.covariances = jnp.array([jnp.eye(self.dim)] * self.n_components)
            elif value.shape == (self.dim, self.dim):
                self.covariances = jnp.repeat(value[jnp.newaxis, :, :], self.n_components, axis=0)

            self.precisions = vmap(jnp.linalg.inv)(self.covariances)
            self.precision_sqrts = vmap(jnp.linalg.cholesky)(self.precisions)
            self.precision_dets = vmap(jnp.linalg.det)(self.precisions)

            self.component_potential = jit(self.component_potential)
            self.component_potentials = jit(vmap(self.component_potential, (None, 0), 0))
            self.component_dens = jit(self.component_dens)
            self.dens = jit(self.dens)
            self.component_potential = jit(self.component_potential)

    @property
    def n_components(self) -> int:
        return len(self.means)

    def component_potential(self,
                            x: jnp.ndarray,
                            component_index: int) -> Union[float, jnp.ndarray]:
        return 0.5 * jnp.sum(jnp.square((x - self.means[component_index]) @
                                      self.precision_sqrts[component_index].T), axis=-1) \
               - jnp.log(self.weights[component_index]
                        * self.precision_dets[component_index]
                        / jnp.power(2 * jnp.pi, self.dim * 0.5))

    def component_dens(self,
                       x: jnp.ndarray,
                       component_index: int) -> Union[float, jnp.ndarray]:
        return jnp.exp(-self.component_potential(x, component_index))

    def dens(self, x: jnp.ndarray) -> Union[float, jnp.ndarray]:
        return jnp.sum(jnp.exp(-self.component_potentials(x, jnp.arange(self.n_components))), axis=0)

    def likelihood_potential(self,
                             x: jnp.ndarray,
                             random_key: jnp.ndarray = None) -> Union[float, jnp.ndarray]:
        return -jnp.log(self.dens(x))


class DoubleWell(Scenario):
    name = "Double Well"

    def __init__(self,
                 dim: int = 1,
                 **kwargs):
        self.dim = dim
        super().__init__(**kwargs)

    def likelihood_potential(self,
                             x: jnp.ndarray,
                             random_key: jnp.ndarray = None) -> Union[float, jnp.ndarray]:
        return jnp.sum(jnp.power(x, 4), axis=-1) / 4. \
               - jnp.sum(jnp.power(x, 2), axis=-1) / 2.


class Rastrigin(Scenario):
    name = "Rastrigin"

    def __init__(self,
                 dim: int = 1,
                 a: float = 1.,
                 **kwargs):
        self.dim = dim
        self.a = a
        super().__init__(**kwargs)

    def likelihood_potential(self,
                             x: jnp.ndarray,
                             random_key: jnp.ndarray = None) -> Union[float, jnp.ndarray]:
        return self.a*self.dim + jnp.sum(x**2 - self.a * jnp.cos(2 * jnp.pi * x), axis=-1)



########################################################################################################################
# Module: ssm/scenarios/lorenz63.py
# Description: Latent Lorenz63 dynamics with Gaussian transition and observation noise.
#
# Web: https://github.com/SamDuffield/mocat
########################################################################################################################

from jax import numpy as jnp, random, jit
from jax.experimental.ode import odeint

from mocat.src.ssm.nonlinear_gaussian import NonLinearGaussian


def lorenz63_dynamics(x: jnp.ndarray,
                      t: float,
                      sigma: float,
                      rho: float,
                      beta: float) -> jnp.ndarray:
    return jnp.array([sigma * (x[1] - x[0]),
                      x[0] * (rho - x[2]) - x[1],
                      x[0] * x[1] - beta * x[2]])


@jit
def lorenz63_integrator(x: jnp.ndarray,
                        delta_t: float,
                        sigma: float,
                        rho: float,
                        beta: float) -> jnp.ndarray:
    return odeint(lorenz63_dynamics, x, jnp.array([0, delta_t]), sigma, rho, beta)[-1]


class Lorenz63(NonLinearGaussian):
    name = 'Lorenz 63'
    dim = 3

    def __init__(self,
                 sigma: float = 10.,
                 rho: float = 28.,
                 beta: float = 8 / 3,
                 **kwargs):
        self.sigma = sigma
        self.rho = rho
        self.beta = beta
        super().__init__(**kwargs)

    def transition_function(self,
                            x_previous: jnp.ndarray,
                            t_previous: float,
                            t_new: float) -> jnp.ndarray:
        return lorenz63_integrator(x_previous, t_new - t_previous, self.sigma, self.rho, self.beta)



########################################################################################################################
# Module: abc/abc.py
# Description: Core objects for ABC, cases where likelihood can't be evaluated but can be simulated.
#
# Web: https://github.com/SamDuffield/mocat
########################################################################################################################

from typing import Union, Tuple

from jax import numpy as jnp, random
from mocat.src.core import cdict, Scenario, impl_checkable, is_implemented
from mocat.src.sample import Sampler


class ABCScenario(Scenario):
    data: jnp.ndarray = None

    def __init__(self,
                 name: str = None,
                 init_grad: bool = False,
                 **kwargs):
        super().__init__(name, init_grad=init_grad, **kwargs)

    @impl_checkable
    def potential(self,
                  x: jnp.ndarray,
                  random_key: jnp.ndarray) -> Union[float, jnp.ndarray]:
        raise TypeError(f'{self.name} abc_scenario potential is intractable')

    @impl_checkable
    def likelihood_potential(self,
                             x: jnp.ndarray,
                             random_key: jnp.ndarray) -> Union[float, jnp.ndarray]:
        raise TypeError(f'{self.name} abc_scenario likelihood_potential is intractable')

    def distance_function(self,
                          simulated_data: jnp.ndarray) -> Union[float, jnp.ndarray]:
        return jnp.sqrt(jnp.square(simulated_data - self.data).sum())


class ABCSampler(Sampler):

    def startup(self,
                abc_scenario: ABCScenario,
                n: int,
                initial_state: cdict,
                initial_extra: cdict,
                startup_correction: bool = True,
                **kwargs) -> Tuple[cdict, cdict]:
        initial_state, initial_extra = super().startup(abc_scenario, n, initial_state, initial_extra,
                                                       **kwargs)

        if not hasattr(initial_state, 'prior_potential') and is_implemented(abc_scenario.prior_potential):
            initial_extra.random_key, subkey = random.split(initial_extra.random_key)
            initial_state.prior_potential = abc_scenario.prior_potential(initial_state.value, subkey)

        if not hasattr(initial_state, 'simulated_data'):
            initial_extra.random_key, subkey = random.split(initial_extra.random_key)
            initial_state.simulated_data = abc_scenario.likelihood_sample(initial_state.value, subkey)

        if not hasattr(initial_state, 'distance'):
            initial_state.distance = abc_scenario.distance_function(initial_state.simulated_data)
        return initial_state, initial_extra






from mocat.src.kernels import Kernel
from mocat.src.kernels import Gaussian
from mocat.src.kernels import PreconditionedGaussian
from mocat.src.kernels import IMQ

from mocat.src.kernels import median_bandwidth_update
from mocat.src.kernels import mean_bandwidth_update




########################################################################################################################
# Module: ssm/tests/test_ssm_linear_gaussian.py
# Description: Tests for linear TemporalGaussian state space models (and Kalman filtering/smoothing).
#
# Web: https://github.com/SamDuffield/mocat
########################################################################################################################

import unittest

from jax import numpy as jnp, random, vmap
import numpy.testing as npt
from mocat.src.tests.test_ssm import TestSSM
from mocat.src.ssm.linear_gaussian.linear_gaussian import LinearGaussian, TimeHomogenousLinearGaussian
from mocat.src.ssm.linear_gaussian.kalman import run_kalman_filter_for_marginals, run_kalman_smoother_for_marginals


def create_eye_lgssm(dim: int):
    return TimeHomogenousLinearGaussian(initial_mean=jnp.zeros(dim),
                                        initial_covariance=jnp.eye(dim),
                                        transition_matrix=jnp.eye(dim),
                                        transition_covariance=jnp.eye(dim),
                                        likelihood_matrix=jnp.eye(dim),
                                        likelihood_covariance=jnp.eye(dim),
                                        name=f'{dim}D-eye LGSSM')


class Test1D(TestSSM):

    ssm_scenario = create_eye_lgssm(1)

    def test_simulate(self):
        super()._test_simulate()

    def test_bootstrap(self):
        super()._test_bootstrap()

    def test_backward(self):
        super()._test_backward()

    def test_online_smoothing_pf(self):
        super()._test_online_smoothing_pf_full()
        super()._test_online_smoothing_pf_rejection()

    def test_online_smoothing_bs(self):
        super()._test_online_smoothing_bs_full()
        super()._test_online_smoothing_bs_rejection()


class Test5D(TestSSM):

    ssm_scenario = create_eye_lgssm(5)

    def test_simulate(self):
        super()._test_simulate()

    def test_bootstrap(self):
        super()._test_bootstrap()

    def test_backward(self):
        super()._test_backward()
    #
    # def test_online_smoothing_pf(self):
    #     super()._test_online_smoothing_pf_full()
    #     super()._test_online_smoothing_pf_rejection()
    #
    # def test_online_smoothing_bs(self):
    #     super()._test_online_smoothing_bs_full()
    #     super()._test_online_smoothing_bs_rejection()


class TestKalman(unittest.TestCase):
    ssm_scenario: LinearGaussian
    t = jnp.arange(20)

    def _test_kalman_filter(self, sds=3):
        if not hasattr(self, 'sim_samps'):
            self.sim_samps = self.ssm_scenario.simulate(self.t, random.PRNGKey(0))
        filter_means, filter_covs = run_kalman_filter_for_marginals(self.ssm_scenario,
                                                                    self.sim_samps.y,
                                                                    self.t)

        filter_sds = jnp.sqrt(vmap(jnp.diag)(filter_covs))
        npt.assert_array_less(self.sim_samps.x, filter_means + sds * filter_sds)
        npt.assert_array_less(filter_means - sds * filter_sds, self.sim_samps.x)

    def _test_kalman_smoother(self, sds=3):
        if not hasattr(self, 'sim_samps'):
            self.sim_samps = self.ssm_scenario.simulate(self.t, random.PRNGKey(0))
        smoother_means, smoother_covs = run_kalman_smoother_for_marginals(self.ssm_scenario,
                                                                          self.sim_samps.y,
                                                                          self.t)
        smoother_sds = jnp.sqrt(vmap(jnp.diag)(smoother_covs))
        npt.assert_array_less(self.sim_samps.x, smoother_means + sds * smoother_sds)
        npt.assert_array_less(smoother_means - sds * smoother_sds, self.sim_samps.x)


class TestKalman1D(TestKalman):
    ssm_scenario = create_eye_lgssm(1)

    def test_kalman_filter(self):
        super()._test_kalman_filter(3)

    def test_kalman_smoother(self):
        super()._test_kalman_smoother(3)


class TestKalman5D(TestKalman):
    ssm_scenario = create_eye_lgssm(5)

    def test_kalman_filter(self):
        super()._test_kalman_filter(4)

    def test_kalman_smoother(self):
        super()._test_kalman_smoother(4)


if __name__ == '__main__':
    unittest.main()



########################################################################################################################
# Module: abc/scenarios/gk.py
# Description: SIR model.
#
# Web: https://github.com/SamDuffield/mocat
########################################################################################################################

from typing import Union, Tuple

from jax import numpy as jnp, random

from mocat.src.abc.abc import ABCScenario
from mocat.src.utils import _while_loop_stacked

exponential_sample_buffer = 1e-6


def sir_single_step(carry: Tuple[jnp.ndarray, float], extra: Tuple[jnp.ndarray, int, jnp.ndarray]):
    si, current_time = carry
    params, population_size, random_key = extra
    t_key, haz_key, random_key = random.split(random_key, 3)

    hazards = jnp.array([params[0] / population_size * si[0] * si[1],
                         params[1] * si[1]])
    sum_hazards = hazards.sum()
    t_increment = -jnp.log(exponential_sample_buffer + random.uniform(t_key)) / sum_hazards
    hazard_probs = hazards / sum_hazards

    possibilities = jnp.array([[si[0] - 1, si[1] + 1],
                               [si[0], si[1] - 1]])
    return (possibilities[random.choice(haz_key, 2, p=hazard_probs)], current_time + t_increment), \
           (params, population_size, random_key)


def sir_simulate(initial_si: jnp.ndarray,
                 params: jnp.ndarray,
                 population_size: int,
                 random_key: jnp.ndarray,
                 max_iter: int = 10000) -> Tuple[jnp.ndarray, jnp.ndarray]:
    (simulated_si, simulated_times), _, _ = _while_loop_stacked(lambda carry, extra: carry[0][1] > 0,
                                                                sir_single_step,
                                                                ((initial_si, 0.),
                                                                 (params, population_size, random_key)),
                                                                max_iter)

    return simulated_times, simulated_si


class SIR(ABCScenario):
    name: str = 'Lotka-Volterra'
    dim: int = 2

    initial_si: jnp.ndarray = None
    prior_rates: jnp.ndarray = jnp.ones(dim)
    times: jnp.ndarray = None
    data: jnp.ndarray = None
    max_iter: int = 10000

    def prior_sample(self,
                     random_key: jnp.ndarray) -> jnp.ndarray:
        return -jnp.log(exponential_sample_buffer + random.uniform(random_key, (self.dim,))) / self.prior_rates

    def prior_potential(self,
                        x: jnp.ndarray,
                        random_key: jnp.ndarray = None) -> Union[float, jnp.ndarray]:
        return (self.prior_rates * x).sum()

    def simulate_times_and_si(self,
                              x: jnp.ndarray,
                              random_key: jnp.ndarray) -> Tuple[jnp.ndarray, jnp.ndarray]:
        return sir_simulate(self.initial_si, x, self.initial_si.sum(),
                            random_key, max_iter=self.max_iter)


class TransformedSIR(SIR):

    @staticmethod
    def constrain(x):
        return jnp.exp(x)

    @staticmethod
    def unconstrain(x):
        return jnp.log(x)

    def prior_sample(self,
                     random_key: jnp.ndarray) -> jnp.ndarray:
        return self.unconstrain(super().prior_sample(random_key))

    def prior_potential(self,
                        x: jnp.ndarray,
                        random_key: jnp.ndarray = None) -> Union[float, jnp.ndarray]:
        return (self.prior_rates * jnp.exp(x) - x).sum()

    def simulate_times_and_si(self,
                              x: jnp.ndarray,
                              random_key: jnp.ndarray) -> Tuple[jnp.ndarray, jnp.ndarray]:
        return super().simulate_times_and_si(self.constrain(x), random_key)



########################################################################################################################
# Module: tests/test_toy_scenarios.py
# Description: Tests for univariate and multivariate toy scenarios
#
# Web: https://github.com/SamDuffield/mocat
########################################################################################################################


import unittest

import jax.numpy as jnp
import numpy.testing as npt

from mocat.src.scenarios import toy_examples


class Test1DGaussian(unittest.TestCase):

    scenario = toy_examples.Gaussian(1, prior_potential=lambda x, rk: 0.)

    def test_basic(self):
        npt.assert_equal(self.scenario.dim, 1)
        npt.assert_equal(self.scenario.mean.shape, (1,))
        npt.assert_equal(self.scenario.covariance.shape, (1,))
        npt.assert_equal(self.scenario.precision_sqrt.shape, (1,))

    def test_scalar_potential(self):
        npt.assert_array_equal(self.scenario.potential(0.), 0.)
        npt.assert_array_equal(self.scenario.potential(5.), 12.5)

    def test_scalar_grad_potential(self):
        npt.assert_array_equal(self.scenario.grad_potential(0.), 0.)
        npt.assert_array_equal(self.scenario.grad_potential(5.), 5.)

    def test_setcov(self):
        self.scenario.covariance = jnp.array([[7.]])
        npt.assert_array_almost_equal(self.scenario.precision_sqrt, 0.37796447)


class TestNDGaussian(unittest.TestCase):

    scenario = toy_examples.Gaussian(10, prior_potential=lambda x, rk: 0.)

    def test_basic(self):
        npt.assert_equal(self.scenario.dim, 10)
        npt.assert_equal(self.scenario.mean.shape, (10,))
        npt.assert_equal(self.scenario.covariance.shape, (10,))
        npt.assert_equal(self.scenario.precision_sqrt.shape, (10,))

    def test_array_potential(self):
        npt.assert_array_equal(self.scenario.potential(jnp.zeros(10)), 0.)
        npt.assert_array_equal(self.scenario.potential(jnp.ones(10) * 3), 45.)

    def test_array_grad_potential(self):
        npt.assert_array_equal(self.scenario.grad_potential(jnp.zeros(10)), jnp.zeros(10))
        npt.assert_array_equal(self.scenario.grad_potential(jnp.ones(10) * 3.), jnp.ones(10) * 3.)


class Test1DGaussianMixture(unittest.TestCase):

    scenario = toy_examples.GaussianMixture(means=jnp.array([0, 1]), prior_potential=lambda x, rk: 0.)

    def test_basic(self):
        npt.assert_equal(self.scenario.dim, 1)
        npt.assert_equal(self.scenario.means.shape, (2, 1))
        npt.assert_equal(self.scenario.covariances.shape, (2, 1, 1))
        npt.assert_equal(self.scenario.precisions.shape, (2, 1, 1))
        npt.assert_equal(self.scenario.precision_sqrts.shape, (2, 1, 1))
        npt.assert_equal(self.scenario.precision_dets.shape, (2,))

    def test_scalar_potential(self):
        npt.assert_array_equal(self.scenario.potential(0.), self.scenario.potential(1.))
        npt.assert_array_almost_equal(self.scenario.potential(5.), 9.601038)

    def test_scalar_grad_potential(self):
        npt.assert_array_equal(self.scenario.grad_potential(0.), -self.scenario.grad_potential(1.))
        npt.assert_array_almost_equal(self.scenario.grad_potential(5.), 4.010987)

    def test_setcovs(self):
        self.scenario.covariances = jnp.repeat(jnp.array([[7.]])[jnp.newaxis, :, :], 2, axis=0)
        npt.assert_array_equal(self.scenario.precisions, jnp.repeat(jnp.array([[1 / 7.]])[jnp.newaxis, :, :], 2, axis=0))
        npt.assert_array_almost_equal(self.scenario.precision_sqrts,
                                      jnp.repeat(jnp.array([[0.37796447]])[jnp.newaxis, :, :], 2, axis=0))


class TestNDGaussianMixture(unittest.TestCase):

    scenario = toy_examples.GaussianMixture(means=jnp.arange(15).reshape(5, 3), prior_potential=lambda x, rk: 0.)

    def test_basic(self):
        npt.assert_equal(self.scenario.dim, 3)
        npt.assert_equal(self.scenario.means.shape, (5, 3))
        npt.assert_equal(self.scenario.covariances.shape, (5, 3, 3))
        npt.assert_equal(self.scenario.precisions.shape, (5, 3, 3))
        npt.assert_equal(self.scenario.precision_sqrts.shape, (5, 3, 3))
        npt.assert_equal(self.scenario.precision_dets.shape, (5,))

    def test_array_potential(self):
        npt.assert_array_almost_equal(self.scenario.potential(jnp.zeros(3)), 6.8662534)
        npt.assert_array_almost_equal(self.scenario.potential(jnp.ones(3) * 2), 6.8552055)

    def test_array_grad_potential(self):
        npt.assert_array_almost_equal(self.scenario.grad_potential(jnp.zeros(3)), jnp.array([0., -1, -2]))
        npt.assert_array_almost_equal(self.scenario.grad_potential(jnp.ones(3) * 2.), jnp.array([1.967039,
                                                                                               0.967039,
                                                                                               -0.032961]))


class Test1DDoubleWell(unittest.TestCase):

    scenario = toy_examples.DoubleWell(1, prior_potential=lambda x, rk: 0.)

    def test_basic(self):
        npt.assert_equal(self.scenario.dim, 1)

    def test_scalar_potential(self):
        npt.assert_array_equal(self.scenario.potential(-3.), self.scenario.potential(3.))
        npt.assert_array_equal(self.scenario.potential(0.), 0)
        npt.assert_array_almost_equal(self.scenario.potential(5.), 143.75)

    def test_scalar_grad_potential(self):
        npt.assert_array_equal(self.scenario.grad_potential(-3.), -self.scenario.grad_potential(3.))
        npt.assert_array_equal(self.scenario.grad_potential(0.), 0.)
        npt.assert_array_equal(self.scenario.grad_potential(1.), 0.)
        npt.assert_array_equal(self.scenario.grad_potential(-1.), 0.)
        npt.assert_array_almost_equal(self.scenario.grad_potential(5.), 120.)


class TestNDDoubleWell(unittest.TestCase):

    scenario = toy_examples.DoubleWell(6, prior_potential=lambda x, rk: 0.)

    def test_basic(self):
        npt.assert_equal(self.scenario.dim, 6)

    def test_array_potential(self):
        npt.assert_array_equal(self.scenario.potential(jnp.zeros(6)), 0.)
        npt.assert_array_equal(self.scenario.potential(jnp.ones(3) * 2), 6.)
        npt.assert_array_almost_equal(self.scenario.potential(jnp.arange(6)), 217.25)

    def test_array_grad_potential(self):
        npt.assert_array_equal(self.scenario.grad_potential(jnp.zeros(6)), jnp.zeros(6))
        npt.assert_array_equal(self.scenario.grad_potential(jnp.ones(6)), jnp.zeros(6))
        npt.assert_array_equal(self.scenario.grad_potential(-jnp.zeros(6)), jnp.zeros(6))
        npt.assert_array_almost_equal(self.scenario.grad_potential(jnp.arange(6, dtype='float32')),
                                      jnp.array([0., 0., 6., 24., 60., 120.]))


if __name__ == '__main__':
    unittest.main()



########################################################################################################################
# Module: utils.py
# Description: Some useful functions including TemporalGaussian potential evaluations and leapfrog integrator.
#
# Web: https://github.com/SamDuffield/mocat
########################################################################################################################

from functools import partial
from typing import Any, Union, Callable, Tuple
from warnings import warn

from jax import jit, numpy as jnp, vmap
from jax.lax import scan, while_loop, cond
from mocat.src.core import cdict


@partial(jit, static_argnums=(0,))
def leave_one_out_indices(n: int,
                          remove_index: int) -> jnp.ndarray:
    full_inds = jnp.arange(n + 1)
    leave_one_out_inds = jnp.where(jnp.arange(n) < remove_index, full_inds[:-1], full_inds[1:])
    return leave_one_out_inds


@jit
def _vectorised_gaussian_potential(x: jnp.ndarray,
                                   mean: Union[float, jnp.ndarray],
                                   sqrt_prec: jnp.ndarray) -> Union[float, jnp.ndarray]:
    diff = x - mean
    return 0.5 * jnp.square(diff @ sqrt_prec).sum(axis=-1)


@jit
def _mv_gaussian_potential(x: jnp.ndarray,
                           mean: Union[jnp.ndarray, float],
                           prec: jnp.ndarray) -> float:
    diff = x - mean
    return 0.5 * diff.T @ prec @ diff


@jit
def _mv_gaussian_potential_diag(x: jnp.ndarray,
                                mean: Union[jnp.ndarray, float],
                                prec_diag: Union[float, jnp.ndarray]) -> Union[float, jnp.ndarray]:
    diff = x - mean
    return 0.5 * jnp.sum(jnp.square(diff) * prec_diag, axis=-1)


def gaussian_potential(x: jnp.ndarray,
                       mean: Union[float, jnp.ndarray] = 0.,
                       prec: Union[float, jnp.ndarray] = None,
                       sqrt_prec: Union[float, jnp.ndarray] = None,
                       det_prec: float = None) -> Union[float, jnp.ndarray]:
    # sqrt_prec such that prec = sqrt_prec @ sqrt_prec.T
    x = jnp.atleast_1d(x)
    d = x.shape[-1]

    if prec is None and sqrt_prec is None:
        prec = 1.

    if isinstance(prec, float) or (prec is not None and prec.ndim == 0):
        prec = jnp.ones(d) * prec

    if isinstance(sqrt_prec, float) or (sqrt_prec is not None and sqrt_prec.ndim == 0):
        sqrt_prec = jnp.ones(d) * sqrt_prec

    if det_prec is None:
        if prec is not None and prec.ndim < 2:
            det_prec = jnp.prod(prec)
        elif sqrt_prec is not None and sqrt_prec.ndim < 2:
            det_prec = jnp.prod(sqrt_prec) ** 2

    if det_prec is None:
        # full precision matrix given but no det - computing without norm constant
        neg_log_z = 0
        warn('gaussian_potential queried with non-diagonal prec (or sqrt-prec) but no det_prec given'
             ' -> executing without normalising constant term')
    else:
        neg_log_z = (d * jnp.log(2 * jnp.pi) - jnp.log(det_prec)) / 2

    if x.ndim == 1 and sqrt_prec is None:
        # Single vals value (not vectorised)
        if prec is None:
            out_val = _mv_gaussian_potential_diag(x, mean, 1.)
        elif prec.ndim < 2:
            out_val = _mv_gaussian_potential_diag(x, mean, prec)
        else:
            out_val = _mv_gaussian_potential(x, mean, prec)
    else:
        # Multiple vals values (vectorised)
        if prec is not None and sqrt_prec is None:
            if prec.ndim < 2:
                sqrt_prec = jnp.sqrt(prec)
            else:
                sqrt_prec = jnp.linalg.cholesky(prec)
                warn('vectorised gaussian_potential queried with prec rather than sqrt_prec'
                     '-> executing using Cholesky decomp')

        if sqrt_prec is None:
            out_val = _mv_gaussian_potential_diag(x, mean, 1.)
        elif sqrt_prec.ndim < 2:
            out_val = _mv_gaussian_potential_diag(x, mean, sqrt_prec ** 2)
        else:
            out_val = _vectorised_gaussian_potential(x, mean, sqrt_prec)
    return out_val + neg_log_z


@partial(jit, static_argnums=(0, 1))
def _leapfrog(prior_potential_and_grad: Callable,
              likelihood_potential_and_grad: Callable,
              state: cdict,
              stepsize: float,
              random_keys: jnp.ndarray,
              temperature: float) -> cdict:
    leapfrog_steps = len(random_keys)

    def leapfrog_step(init_state: cdict,
                      i: int):
        new_state = init_state.copy()

        p_half = init_state.momenta - stepsize / 2. * init_state.grad_potential

        new_state.value = init_state.value + stepsize * p_half

        new_state.prior_potential, new_state.grad_prior_potential \
            = prior_potential_and_grad(new_state.value, random_keys[i])

        new_state.likelihood_potential, new_state.grad_likelihood_potential \
            = likelihood_potential_and_grad(new_state.value, random_keys[i])

        new_state.potential = new_state.prior_potential + temperature * new_state.likelihood_potential
        new_state.grad_potential = new_state.grad_prior_potential + temperature * new_state.grad_likelihood_potential

        new_state.momenta = p_half - stepsize / 2. * new_state.grad_potential

        next_sample_chain = new_state.copy()
        next_sample_chain.momenta = jnp.vstack([p_half, new_state.momenta])
        return new_state, next_sample_chain

    final_leapfrog, all_leapfrog = scan(leapfrog_step, state, jnp.arange(leapfrog_steps))

    all_leapfrog.momenta = jnp.concatenate(all_leapfrog.momenta)

    all_leapfrog = state[jnp.newaxis] + all_leapfrog

    return all_leapfrog


def leapfrog(prior_potential_and_grad: Callable,
             likelihood_potential_and_grad: Callable,
             state: cdict,
             stepsize: float,
             random_keys: jnp.ndarray,
             temperature: float = 1.) -> cdict:
    return _leapfrog(prior_potential_and_grad, likelihood_potential_and_grad,
                     state, stepsize, random_keys, temperature)


@partial(jit, static_argnums=(0, 1, 3))
def _while_loop_stacked(cond_fun: Callable,
                        body_fun: Callable,
                        init_carry: Tuple[Any, Any],
                        max_iter: int) -> Tuple[Any, Tuple[Any, Any], int]:
    # init_carry = (init_state, extra)
    # cond_func: (state, extra) -> bool
    # body_func: (state, extra) -> (state, extra)
    # while_loop_stacked -> stacked(state)

    def update_cond(previous_carry: Any,
                    final_iter: int,
                    live_iter: int) -> bool:
        return cond(live_iter <= (final_iter + 1),
                    lambda args: cond_fun(*args),
                    lambda _: False,
                    previous_carry)

    def update_kernel(previous_val_and_final_iter: Tuple[Any, int],
                      live_iter: int) -> Tuple[Tuple[Any, int], Any]:
        previous_val, final_iter = previous_val_and_final_iter

        update_bool = update_cond(previous_val, final_iter, live_iter)

        final_iter = jnp.where(update_bool, live_iter, final_iter)

        new_val = cond(update_bool,
                       lambda args: body_fun(*args),
                       lambda _: init_carry,
                       previous_val)

        return (new_val, final_iter), new_val[0]

    final_val_and_final_iter, stack = scan(update_kernel, (init_carry, 0), jnp.arange(1, max_iter + 1))
    final_val, final_final_iter = final_val_and_final_iter
    return stack, final_val, final_final_iter


def while_loop_stacked(cond_fun: Callable,
                       body_fun: Callable,
                       init_carry: Tuple[Any, Any],
                       max_iter: int = 1000) -> Any:
    full_stack, final_carry, final_int = _while_loop_stacked(cond_fun, body_fun, init_carry, max_iter)
    return tuple(a[:final_int] for a in full_stack) if isinstance(full_stack, tuple) else full_stack[:final_int]


@partial(jit, static_argnums=(0,))
def bisect(fun: Callable,
           bounds: Union[list, jnp.ndarray],
           max_iter: int = 1000,
           tol: float = 1e-5) -> Tuple[jnp.ndarray, jnp.ndarray, int]:
    evals = jnp.array([fun(bounds[0]), fun(bounds[1])])
    increasing_bool = evals[1] > evals[0]

    def conv_check(int_state: Tuple[jnp.ndarray, jnp.ndarray, int]) -> bool:
        int_bounds, int_evals, iter_ind = int_state
        return ~jnp.any(jnp.array([jnp.min(jnp.abs(int_evals)) < tol,
                                   iter_ind >= max_iter,
                                   jnp.all(int_evals < 0),
                                   jnp.all(int_evals > 0)]))

    def body_func(int_state: Tuple[jnp.ndarray, jnp.ndarray, int]) -> Tuple[jnp.ndarray, jnp.ndarray, int]:
        int_bounds, int_evals, iter_ind = int_state

        new_pos = int_bounds[0] - int_evals[0] * (int_bounds[1] - int_bounds[0]) / (int_evals[1] - int_evals[0])
        new_eval = fun(new_pos)

        replace_upper = jnp.where(increasing_bool, new_eval > 0, new_eval < 0)

        out_bounds = jnp.where(replace_upper, jnp.array([int_bounds[0], new_pos]), jnp.array([new_pos, int_bounds[1]]))
        out_evals = jnp.where(replace_upper, jnp.array([int_evals[0], new_eval]), jnp.array([new_eval, int_evals[1]]))

        return out_bounds, out_evals, iter_ind + 1

    fin_bounds, fin_evals, fin_iter = while_loop(conv_check,
                                                 body_func,
                                                 (bounds, evals, 0))

    return fin_bounds, fin_evals, fin_iter


def extract_dimension(*args):
    for a in args:
        if isinstance(a, jnp.ndarray) and a.ndim > 0:
            return a.shape[-1]
    return None


def reset_covariance(obj: Any,
                     key: str,
                     value: Any):
    prec_key = key.replace('covariance', 'precision')
    if value.ndim < 2:
        sqrt_val = jnp.sqrt(value)
        setattr(obj, key + '_sqrt', sqrt_val)
        setattr(obj, prec_key + '_sqrt', 1 / sqrt_val)
        setattr(obj, prec_key + '_det', 1 / value)
        setattr(obj, prec_key + '_mul', lambda a, b: a * b)
    else:
        sqrt_mat = jnp.linalg.cholesky(value)
        setattr(obj, key + '_sqrt', sqrt_mat)
        setattr(obj, prec_key + '_sqrt', jnp.linalg.inv(sqrt_mat))
        setattr(obj, prec_key + '_det', 1 / jnp.linalg.det(value))
        setattr(obj, prec_key + '_mul', lambda a, b: a @ b)


# @jit
# def _bfgs_sqrt_pqut(vals: jnp.ndarray,
#                     grads: jnp.ndarray,
#                     init_hessian_sqrt_diag: jnp.ndarray,
#                     r: float,
#                     update_bools: jnp.ndarray) -> Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray, jnp.ndarray]:
#     s = vals[1:] - vals[:-1]
#     y = grads[1:] - grads[:-1]
#
#     m, d = s.shape
#
#     update_bools = jnp.ones(m) * update_bools
#
#     def body_fun(val, i):
#         us, ts = val
#         sit_yi = jnp.dot(s[i], y[i])
#         sit_si = jnp.dot(s[i], s[i])
#         sit_Binit_si = jnp.square(s[i] * init_hessian_sqrt_diag).sum()
#         #
#         # thetai = (1 - r) * sit_si / \
#         #          (sit_Binit_si - jnp.dot(init_hessian_sqrt_diag * s[i], init_hessian_sqrt_diag * y[i]))
#         thetai = (sit_si - r * sit_Binit_si) / (sit_si - sit_yi)
#         ybar_i = jnp.where(sit_yi < (r * sit_Binit_si), thetai * y[i] + (1 - thetai) * s[i], y[i])
#         sit_yi = jnp.dot(s[i], ybar_i)
#
#         Ct_si = _bfgs_sqrt_transpose_prod(us, ts, s[i], init_hessian_sqrt_diag)
#         B_si = _bfgs_sqrt_prod(us, ts, Ct_si, init_hessian_sqrt_diag)
#         sit_B_si = jnp.dot(s[i], B_si)
#
#         min_denoms = jnp.minimum(jnp.abs(sit_B_si), jnp.abs(sit_yi))
#         update_bool = jnp.logical_or(r == -jnp.inf, sit_yi > 0)
#         update_bool = jnp.logical_and(min_denoms > 1e-10, update_bool)
#         update_bool = jnp.logical_and(update_bools[i], update_bool)
#
#         p = jnp.where(update_bool, s[i] / sit_yi, jnp.zeros(d))
#         q = jnp.where(update_bool, jnp.sqrt(sit_yi / sit_B_si) * B_si + ybar_i, jnp.zeros(d))
#         u = jnp.where(update_bool, jnp.sqrt(sit_B_si / sit_yi) * ybar_i + B_si, jnp.zeros(d))
#         t = jnp.where(update_bool, s[i] / sit_B_si, jnp.zeros(d))
#         return (index_update(us, i, u), index_update(ts, i, t)), (p, q, u, t)
#
#     return scan(body_fun, (jnp.zeros((m, d)), jnp.zeros((m, d))), jnp.arange(m))[1]
#
#
# def bfgs_sqrt_pqut(vals: jnp.ndarray,
#                    grads: jnp.ndarray,
#                    init_hessian_sqrt_diag: jnp.ndarray = 1.,
#                    r: float = -jnp.inf,
#                    update_bools: jnp.ndarray = 1) -> Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray, jnp.ndarray]:
#     # Uses ps, qs for Hessian inverse approximation and us, ts for Hessian approximation
#     # 0 < r < 1 enforces positive definiteness
#     return _bfgs_sqrt_pqut(vals, grads, init_hessian_sqrt_diag, r, update_bools)


@jit
def _bfgs_sqrt_pqut(vals: jnp.ndarray,
                    grads: jnp.ndarray,
                    init_hessian_sqrt_diag: jnp.ndarray,
                    r: float,
                    update_bools: jnp.ndarray) -> Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray, jnp.ndarray]:
    s = vals[1:] - vals[:-1]
    y = grads[1:] - grads[:-1]

    sit_Binit_si_all = jnp.square(s * init_hessian_sqrt_diag).sum(-1)
    sty = (s * y).sum(-1)
    rats = jnp.where(sit_Binit_si_all < 1e-10, -jnp.inf, -sty/sit_Binit_si_all)
    lambd = jnp.maximum(0., rats.max() + r)
    lambd = lambd * init_hessian_sqrt_diag ** 2

    y = y + lambd * s

    m, d = s.shape

    update_bools = jnp.ones(m) * update_bools

    def body_fun(val, i):
        us, ts = val
        sit_yi = jnp.dot(s[i], y[i])
        sit_si = jnp.dot(s[i], s[i])
        sit_Binit_si = jnp.square(s[i] * init_hessian_sqrt_diag).sum()

        sit_yi = jnp.dot(s[i], y[i])

        Ct_si = _bfgs_sqrt_transpose_prod(us, ts, s[i], init_hessian_sqrt_diag)
        B_si = _bfgs_sqrt_prod(us, ts, Ct_si, init_hessian_sqrt_diag)
        sit_B_si = jnp.dot(s[i], B_si)

        min_denoms = jnp.minimum(jnp.abs(sit_B_si), jnp.abs(sit_yi))
        update_bool = jnp.logical_or(r == -jnp.inf, sit_yi > 0)
        update_bool = jnp.logical_and(min_denoms > 1e-10, update_bool)
        update_bool = jnp.logical_and(update_bools[i], update_bool)

        p = jnp.where(update_bool, s[i] / sit_yi, jnp.zeros(d))
        q = jnp.where(update_bool, jnp.sqrt(sit_yi / sit_B_si) * B_si + y[i], jnp.zeros(d))
        u = jnp.where(update_bool, jnp.sqrt(sit_B_si / sit_yi) * y[i] + B_si, jnp.zeros(d))
        t = jnp.where(update_bool, s[i] / sit_B_si, jnp.zeros(d))
        return (us.at[i].set(u), ts.at[i].set(t)), (p, q, u, t)

    return scan(body_fun, (jnp.zeros((m, d)), jnp.zeros((m, d))), jnp.arange(m))[1]


def bfgs_sqrt_pqut(vals: jnp.ndarray,
                   grads: jnp.ndarray,
                   init_hessian_sqrt_diag: jnp.ndarray = 1.,
                   r: float = -jnp.inf,
                   update_bools: jnp.ndarray = 1) -> Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray, jnp.ndarray]:
    # Uses ps, qs for Hessian inverse approximation and us, ts for Hessian approximation
    # 0 < r enforces positive definiteness
    return _bfgs_sqrt_pqut(vals, grads, init_hessian_sqrt_diag, r, update_bools)


@jit
def _bfgs_sqrt_prod(ps: jnp.ndarray,
                    qs: jnp.ndarray,
                    z: jnp.ndarray,
                    init_diag: jnp.ndarray) -> jnp.ndarray:
    def body_fun(ci_z, i):
        return ci_z - ps[i] * jnp.dot(qs[i], ci_z), None

    return scan(body_fun, init_diag * z, jnp.arange(len(ps)))[0]


def bfgs_sqrt_prod(ps: jnp.ndarray,
                   qs: jnp.ndarray,
                   z: jnp.ndarray,
                   init_diag: jnp.ndarray = 1.) -> jnp.ndarray:
    return _bfgs_sqrt_prod(ps, qs, z, init_diag)


@jit
def _bfgs_sqrt_transpose_prod(ps: jnp.ndarray,
                              qs: jnp.ndarray,
                              z: jnp.ndarray,
                              init_diag: jnp.ndarray) -> jnp.ndarray:
    def body_fun(ci_z, i):
        return ci_z - qs[i] * jnp.dot(ps[i], ci_z), None

    return init_diag * scan(body_fun, z, jnp.arange(len(ps) - 1, -1, -1))[0]


def bfgs_sqrt_transpose_prod(ps: jnp.ndarray,
                             qs: jnp.ndarray,
                             z: jnp.ndarray,
                             init_diag: jnp.ndarray = 1.) -> jnp.ndarray:
    return _bfgs_sqrt_transpose_prod(ps, qs, z, init_diag)


@jit
def _bfgs_sqrt_det(ps: jnp.ndarray,
                   qs: jnp.ndarray,
                   init_diag: jnp.ndarray) -> jnp.ndarray:
    return jnp.prod(init_diag) * (1 - vmap(jnp.dot)(ps, qs)).prod()


def bfgs_sqrt_det(ps: jnp.ndarray,
                  qs: jnp.ndarray,
                  init_diag: jnp.ndarray = 1.) -> jnp.ndarray:
    return _bfgs_sqrt_det(ps, qs, init_diag)


@jit
def l2_distance_matrix(vals: jnp.ndarray) -> jnp.ndarray:
    return vmap(lambda x: vmap(lambda y: jnp.sum(jnp.square(x - y)))(vals))(vals) ** 0.5


def _gc_close(z_over_rad):
    return -0.25 * z_over_rad ** 5 \
           + 0.5 * z_over_rad ** 4 \
           + 5 / 8 * z_over_rad ** 3 \
           - 5 / 3 * z_over_rad ** 2 \
           + 1


def _gc_mid(z_over_rad):
    return 1 / 12 * z_over_rad ** 5 \
           - 0.5 * z_over_rad ** 4 \
           + 5 / 8 * z_over_rad ** 3 \
           + 5 / 3 * z_over_rad ** 2 \
           - 5 * z_over_rad \
           + 4 \
           - 2 / 3 / z_over_rad


def _gc(z, radius):
    z_over_rad = jnp.abs(z) / radius
    out = jnp.zeros_like(z)
    out = jnp.where(z < radius, _gc_close(z_over_rad), out)
    out = jnp.where((z >= radius) * (z < 2 * radius), _gc_mid(z_over_rad), out)
    return out


def gc_matrix(dim: int,
              radius: float):
    return vmap(lambda i: _gc(jnp.min(jnp.array([jnp.abs(jnp.arange(dim) - i),
                                                 jnp.abs(jnp.arange(dim) + dim - i),
                                                 jnp.abs(jnp.arange(dim) - dim - i)]), axis=0), radius)
                )(jnp.arange(dim))


@jit
def kalman_gain(cov_sqrt: jnp.ndarray,
                lik_mat: jnp.ndarray,
                lik_cov_sqrt: jnp.ndarray,
                inv_mat: jnp.ndarray = None) -> jnp.ndarray:
    if inv_mat is None:
        lik_mat_cov_sqrt = lik_mat @ cov_sqrt
        inv_mat = jnp.linalg.inv(lik_mat_cov_sqrt @ lik_mat_cov_sqrt.T + lik_cov_sqrt @ lik_cov_sqrt.T)
    return cov_sqrt @ cov_sqrt.T @ lik_mat.T @ inv_mat



########################################################################################################################
# Module: mcmc/qn_underdamped.py
# Description: Underdamped Langevin kernel with L-BFGS inverse Hessian pre-conditioner
#
# Web: https://github.com/SamDuffield/mocat
########################################################################################################################

from functools import partial
from typing import Tuple, Union, Callable

import jax.numpy as jnp
from jax import random, jit, vmap
from jax.lax import scan

from mocat.src.core import cdict, Scenario
from mocat.src.mcmc.sampler import MCMCSampler
from mocat.src.mcmc.metropolis import Metropolis
from mocat.src.utils import bfgs_sqrt_pqut, bfgs_sqrt_prod, bfgs_sqrt_transpose_prod
from mocat.src.transport.smc import RMMetropolisedSMCSampler


# Underdamped Langevin
# L-BFGS approximated inverse Hessian pre-conditioner/mass matrix
class QNUnderdamped(MCMCSampler):
    name = 'Quasi-Newton Underdamped'
    correction = Metropolis

    def __init__(self,
                 stepsize: float = None,
                 leapfrog_steps: int = 1,
                 friction: float = 1.0,
                 m: int = 30,
                 init_hessian_sqrt_diag: float = 1.,
                 pos_def_r: float = 0.5,
                 sort_bfgs: bool = False,
                 max_val: float = 1e10):
        super().__init__()
        self.parameters.stepsize = stepsize
        self.parameters.leapfrog_steps = leapfrog_steps
        self.parameters.friction = friction
        self.parameters.m = m
        self.parameters.init_hessian_sqrt_diag = init_hessian_sqrt_diag
        self.parameters.pos_def_r = pos_def_r
        self.parameters.max_val = max_val
        self.sort_bfgs = sort_bfgs
        self.tuning.target = 0.651

    def startup(self,
                scenario: Scenario,
                n: int,
                initial_state: cdict,
                initial_extra: cdict,
                **kwargs) -> Tuple[cdict, cdict]:
        initial_state, initial_extra = super().startup(scenario, n,
                                                       initial_state, initial_extra, **kwargs)
        initial_extra.random_key, prior_key, likelihood_key = random.split(initial_extra.random_key, 3)
        if not hasattr(initial_state, 'prior_potential') or not hasattr(initial_state, 'grad_prior_potential'):
            initial_state.prior_potential, initial_state.grad_prior_potential \
                = scenario.prior_potential_and_grad(initial_state.value, prior_key)
        if not hasattr(initial_state, 'likelihood_potential') or not hasattr(initial_state,
                                                                             'grad_likelihood_potential'):
            initial_state.likelihood_potential, initial_state.grad_likelihood_potential \
                = scenario.likelihood_potential_and_grad(initial_state.value, likelihood_key)
        initial_state.potential = initial_state.prior_potential \
                                  + scenario.temperature * initial_state.likelihood_potential
        initial_state.grad_potential = initial_state.grad_prior_potential \
                                       + scenario.temperature * initial_state.grad_likelihood_potential

        if not hasattr(initial_state, 'momenta') or initial_state.momenta.shape[-1] != scenario.dim:
            initial_state.momenta = jnp.zeros(scenario.dim)

        if self.sort_bfgs:
            self._bfgs_vals_grads = self._bfgs_vals_grads_sort
        else:
            self._bfgs_vals_grads = lambda v, p, g: (v, g)

        if not hasattr(initial_state, 'memory'):
            initial_state.memory \
                = cdict(value=jnp.zeros((initial_extra.parameters.m, scenario.dim)),
                        prior_potential=jnp.zeros(initial_extra.parameters.m) - jnp.inf,
                        grad_prior_potential=jnp.zeros((initial_extra.parameters.m, scenario.dim)),
                        likelihood_potential=jnp.zeros(initial_extra.parameters.m) - jnp.inf,
                        grad_likelihood_potential=jnp.zeros((initial_extra.parameters.m, scenario.dim)),
                        init_hessian_sqrt_diag=initial_extra.parameters.init_hessian_sqrt_diag)
            initial_state.memory \
                = self.update_memory(initial_state.memory, jnp.zeros(scenario.dim),
                                     -jnp.inf, jnp.zeros(scenario.dim),
                                     -jnp.inf, jnp.zeros(scenario.dim),
                                     scenario.temperature)

        return initial_state, initial_extra

    @staticmethod
    def _bfgs_vals_grads(value: jnp.ndarray,
                         potential: float,
                         grad_potential: jnp.ndarray) -> Tuple[jnp.ndarray, jnp.ndarray]:
        raise NotImplementedError

    @staticmethod
    def _bfgs_vals_grads_sort(value: jnp.ndarray,
                              potential: jnp.ndarray,
                              grad_potential: jnp.ndarray) -> Tuple[jnp.ndarray, jnp.ndarray]:
        inds = jnp.argsort(potential)
        return value[inds], grad_potential[inds]

    def update_memory(self,
                      memory: cdict,
                      value: jnp.ndarray,
                      prior_potential: float,
                      grad_prior_potential: jnp.ndarray,
                      likelihood_potential: float,
                      grad_likelihood_potential: jnp.ndarray,
                      temperature: float) -> cdict:
        memory.value = jnp.append(memory.value[1:], value[jnp.newaxis], axis=0)
        memory.prior_potential = jnp.append(memory.prior_potential[1:], prior_potential)
        memory.grad_prior_potential \
            = jnp.append(memory.grad_prior_potential[1:], grad_prior_potential[jnp.newaxis], axis=0)
        memory.likelihood_potential = jnp.append(memory.likelihood_potential[1:], likelihood_potential)
        memory.grad_likelihood_potential \
            = jnp.append(memory.grad_likelihood_potential[1:], grad_likelihood_potential[jnp.newaxis], axis=0)

        value, grad_potential \
            = self._bfgs_vals_grads(memory.value,
                                    memory.prior_potential + temperature * memory.likelihood_potential,
                                    memory.grad_prior_potential + temperature * memory.grad_likelihood_potential)

        update_bools = jnp.array(memory.prior_potential[:-1] > -jnp.inf)

        memory.ps, memory.qs, memory.us, memory.ts \
            = bfgs_sqrt_pqut(value, grad_potential,
                             init_hessian_sqrt_diag=memory.init_hessian_sqrt_diag, r=self.parameters.pos_def_r,
                             update_bools=update_bools)
        return memory

    def always(self,
               scenario: Scenario,
               reject_state: cdict,
               reject_extra: cdict) -> Tuple[cdict, cdict]:
        d = scenario.dim

        stepsize = reject_extra.parameters.stepsize
        friction = reject_extra.parameters.friction

        random_keys = random.split(reject_extra.random_key, 2)

        reject_extra.random_key = random_keys[-1]

        reject_state.memory = self.update_memory(reject_state.memory.copy(),
                                                 reject_state.value,
                                                 reject_state.prior_potential,
                                                 reject_state.grad_prior_potential,
                                                 reject_state.likelihood_potential,
                                                 reject_state.grad_likelihood_potential,
                                                 scenario.temperature)

        reject_state.momenta = - reject_state.momenta

        # Update p - exactly according to solution of OU process
        # Accepted even if leapfrog step is rejected
        precon_norm = bfgs_sqrt_prod(reject_state.memory.us, reject_state.memory.ts,
                                     random.normal(random_keys[-2], (d,)),
                                     reject_state.memory.init_hessian_sqrt_diag)
        reject_state.momenta = reject_state.momenta * jnp.exp(- friction * stepsize) \
                               + jnp.sqrt(1 - jnp.exp(- 2 * friction * stepsize)) * precon_norm

        return reject_state, reject_extra

    def proposal(self,
                 scenario: Scenario,
                 reject_state: cdict,
                 reject_extra: cdict) -> Tuple[cdict, cdict]:
        proposed_state = reject_state.copy()

        random_keys = random.split(reject_extra.random_key, self.parameters.leapfrog_steps + 1)
        reject_extra.random_key = random_keys[-1]

        all_leapfrog = bfgs_leapfrog(scenario.prior_potential_and_grad,
                                     scenario.likelihood_potential_and_grad,
                                     proposed_state,
                                     reject_extra.parameters.stepsize,
                                     random_keys[:self.parameters.leapfrog_steps],
                                     proposed_state.memory.ps,
                                     proposed_state.memory.qs,
                                     1 / proposed_state.memory.init_hessian_sqrt_diag,
                                     scenario.temperature)

        proposed_state.momenta = - all_leapfrog.momenta[-1]

        proposed_state.value = all_leapfrog.value[-1]
        proposed_state.prior_potential = all_leapfrog.prior_potential[-1]
        proposed_state.grad_prior_potential = all_leapfrog.grad_prior_potential[-1]
        proposed_state.likelihood_potential = all_leapfrog.likelihood_potential[-1]
        proposed_state.grad_likelihood_potential = all_leapfrog.grad_likelihood_potential[-1]
        proposed_state.potential \
            = proposed_state.prior_potential + scenario.temperature * proposed_state.likelihood_potential
        proposed_state.grad_potential \
            = proposed_state.grad_prior_potential + scenario.temperature * proposed_state.grad_likelihood_potential

        return proposed_state, reject_extra

    def acceptance_probability(self,
                               scenario: Scenario,
                               reject_state: cdict, reject_extra: cdict,
                               proposed_state: cdict, proposed_extra: cdict) -> Union[float, jnp.ndarray]:
        rej_sqrtprec_momenta_diff = bfgs_sqrt_transpose_prod(reject_state.memory.ps, reject_state.memory.qs,
                                                             reject_state.momenta,
                                                             1 / reject_state.memory.init_hessian_sqrt_diag)
        rej_momenta_pot = 0.5 * jnp.square(rej_sqrtprec_momenta_diff).sum()

        prop_sqrtprec_momenta_diff = bfgs_sqrt_transpose_prod(reject_state.memory.ps, reject_state.memory.qs,
                                                              proposed_state.momenta,
                                                              1 / reject_state.memory.init_hessian_sqrt_diag)
        prop_momenta_pot = 0.5 * jnp.square(prop_sqrtprec_momenta_diff).sum()

        pre_min_alpha = jnp.exp(- proposed_state.potential
                                + reject_state.potential
                                - prop_momenta_pot
                                + rej_momenta_pot)
        alpha = jnp.minimum(1., pre_min_alpha)
        alpha = jnp.where(
            jnp.any(jnp.abs(jnp.append(proposed_state.value, proposed_state.momenta))
                    > reject_extra.parameters.max_val), 0., alpha)
        return alpha


@partial(jit, static_argnums=(0, 1))
def _bfgs_leapfrog(prior_potential_and_grad: Callable,
                   likelihood_potential_and_grad: Callable,
                   state: cdict,
                   stepsize: float,
                   random_keys: jnp.ndarray,
                   bfgs_ps: jnp.ndarray,
                   bfgs_qs: jnp.ndarray,
                   bfgs_init_invhessian_sqrt_diag: Union[float, jnp.ndarray],
                   temperature: float) -> cdict:
    leapfrog_steps = len(random_keys)

    def leapfrog_step(init_state: cdict,
                      i: int):
        new_state = init_state.copy()

        p_half = init_state.momenta - stepsize / 2. * init_state.grad_potential
        minv_p_half = bfgs_sqrt_transpose_prod(bfgs_ps, bfgs_qs, p_half, bfgs_init_invhessian_sqrt_diag)
        minv_p_half = bfgs_sqrt_prod(bfgs_ps, bfgs_qs, minv_p_half, bfgs_init_invhessian_sqrt_diag)

        new_state.value = init_state.value + stepsize * minv_p_half

        new_state.prior_potential, new_state.grad_prior_potential \
            = prior_potential_and_grad(new_state.value, random_keys[i])

        new_state.likelihood_potential, new_state.grad_likelihood_potential \
            = likelihood_potential_and_grad(new_state.value, random_keys[i])

        new_state.potential = new_state.prior_potential + temperature * new_state.likelihood_potential
        new_state.grad_potential = new_state.grad_prior_potential + temperature * new_state.grad_likelihood_potential

        new_state.momenta = p_half - stepsize / 2. * new_state.grad_potential

        next_sample_chain = new_state.copy()
        next_sample_chain.momenta = jnp.vstack([p_half, new_state.momenta])
        return new_state, next_sample_chain

    final_leapfrog, all_leapfrog = scan(leapfrog_step, state, jnp.arange(leapfrog_steps))

    all_leapfrog.momenta = jnp.concatenate(all_leapfrog.momenta)

    all_leapfrog = state[jnp.newaxis] + all_leapfrog

    return all_leapfrog


def bfgs_leapfrog(prior_potential_and_grad: Callable,
                  likelihood_potential_and_grad: Callable,
                  state: cdict,
                  stepsize: float,
                  random_keys: jnp.ndarray,
                  bfgs_ps: jnp.ndarray,
                  bfgs_qs: jnp.ndarray,
                  bfgs_init_invhessian_sqrt_diag: Union[float, jnp.ndarray] = 1.,
                  temperature: float = 1.) -> cdict:
    return _bfgs_leapfrog(prior_potential_and_grad, likelihood_potential_and_grad,
                          state, stepsize, random_keys, bfgs_ps, bfgs_qs, bfgs_init_invhessian_sqrt_diag, temperature)



########################################################################################################################
# Module: ssm/nonlinear_gaussian.py
# Description: State-space models with non-linear transition function, linear likelihood
#              and all (time-homogenous) Gaussian noise
#
# Web: https://github.com/SamDuffield/mocat
########################################################################################################################

from functools import partial
from typing import Union, Any, Tuple

from jax import numpy as jnp, random, vmap, jit

from mocat.src.ssm.ssm import StateSpaceModel
from mocat.src.utils import gaussian_potential, extract_dimension, reset_covariance, kalman_gain
from mocat.src.ssm.filtering import ParticleFilter


class NonLinearGaussian(StateSpaceModel):
    # x_{t_new} ~ N(f(x_{t_previous}, t_previous, t_new), Q)
    # y_t ~ N(H @ x_t, R)
    # where Q, H, R are all time-homogenous
    # typically this means t is of the form jnp.arange(start, end, step)

    initial_covariance_sqrt: jnp.ndarray = None
    initial_precision_sqrt: jnp.ndarray = None
    initial_precision_det: float = None
    transition_covariance_sqrt: jnp.ndarray = None
    transition_precision_sqrt: jnp.ndarray = None
    transition_precision_det: float = None
    likelihood_covariance_sqrt: jnp.ndarray = None
    likelihood_precision_sqrt: jnp.ndarray = None
    likelihood_precision_det: jnp.ndarray = None

    def __init__(self,
                 initial_mean: jnp.ndarray = None,
                 initial_covariance: float = None,
                 transition_covariance: jnp.ndarray = None,
                 likelihood_matrix: jnp.ndarray = None,
                 likelihood_covariance: jnp.ndarray = None,
                 name: str = None):
        if self.dim is None:
            self.dim = extract_dimension(initial_mean, initial_covariance,
                                         transition_covariance,
                                         likelihood_matrix, likelihood_covariance)
        if self.dim is None:
            raise AttributeError(f'Could not find dimension for {self.__class__.__name__}')

        if initial_mean is None:
            initial_mean = jnp.zeros(self.dim)
        self.initial_mean = initial_mean

        if initial_covariance is None:
            initial_covariance = jnp.eye(self.dim)
        self.initial_covariance = initial_covariance

        if transition_covariance is None:
            transition_covariance = jnp.eye(self.dim)
        self.transition_covariance = transition_covariance

        if likelihood_matrix is None:
            likelihood_matrix = jnp.eye(self.dim, dtype='float32')

        self.likelihood_matrix = likelihood_matrix
        self.dim_obs = self.likelihood_matrix.shape[0]

        if likelihood_covariance is None:
            likelihood_covariance = jnp.eye(self.dim_obs)
        self.likelihood_covariance = likelihood_covariance

        super().__init__(name=name)

    def __setattr__(self,
                    key: str,
                    value: Any):
        self.__dict__[key] = value
        if key[-10:] == 'covariance':
            reset_covariance(self, key, value)

    def initial_potential(self,
                          x: jnp.ndarray,
                          t: float) -> Union[float, jnp.ndarray]:
        return gaussian_potential(x, self.initial_mean,
                                  sqrt_prec=self.initial_precision_sqrt, det_prec=self.initial_precision_det)

    def initial_sample(self,
                       t: float,
                       random_key: jnp.ndarray) -> Union[float, jnp.ndarray]:
        return random.normal(random_key, shape=(self.dim,)) @ self.initial_covariance_sqrt.T \
               + self.initial_mean

    def transition_function(self,
                            x_previous: jnp.ndarray,
                            t_previous: float,
                            t_new: float) -> jnp.ndarray:
        raise NotImplementedError(f'{self.__class__.__name__} nonlinear transition_function not implemented')

    def transition_potential(self,
                             x_previous: jnp.ndarray,
                             t_previous: float,
                             x_new: jnp.ndarray,
                             t_new: float) -> Union[float, jnp.ndarray]:
        return gaussian_potential(x_new,
                                  self.transition_function(x_previous, t_previous, t_new),
                                  sqrt_prec=self.transition_precision_sqrt, det_prec=self.transition_precision_det)

    def transition_sample(self,
                          x_previous: jnp.ndarray,
                          t_previous: float,
                          t_new: float,
                          random_key: jnp.ndarray) -> jnp.ndarray:
        return self.transition_function(x_previous, t_previous, t_new) \
               + random.normal(random_key, shape=x_previous.shape) @ self.transition_covariance_sqrt

    def likelihood_potential(self,
                             x: jnp.ndarray,
                             y: jnp.ndarray,
                             t: float) -> Union[float, jnp.ndarray]:
        return gaussian_potential(y,
                                  x @ self.likelihood_matrix.T,
                                  sqrt_prec=self.likelihood_precision_sqrt, det_prec=self.likelihood_precision_det)

    def likelihood_sample(self,
                          x: jnp.ndarray,
                          t: float,
                          random_key: jnp.ndarray) -> jnp.ndarray:

        rand_shape = list(x.shape)
        rand_shape[-1] = self.likelihood_covariance_sqrt.shape[0]
        return x @ self.likelihood_matrix.T \
               + random.normal(random_key, shape=rand_shape) @ self.likelihood_covariance_sqrt.T


class OptimalNonLinearGaussianParticleFilter(ParticleFilter):
    # NOTE we assume time-homogenous transition noise covariance
    # that is x_t ~ N(f_t(x_s), Q) where Q is independent of s, t or t-s
    # time-inhomogoneous transition noise covariance would require matrix inversions
    # at every time point rather than just once
    # typically this means t is of the form jnp.arange(start, end, step)

    name = 'Optimal Non-linear Gaussian Particle Filter'

    initial_kalman_gain: jnp.ndarray = None
    initial_conditioned_precision_sqrt: jnp.ndarray = None
    initial_conditioned_covariance_sqrt: jnp.ndarray = None
    initial_conditioned_precision_det: float = None
    proposal_kalman_gain: jnp.ndarray = None
    proposal_precision_sqrt: jnp.ndarray = None
    proposal_covariance_sqrt: jnp.ndarray = None
    proposal_precision_det: float = None
    weight_precision_sqrt: jnp.ndarray = None
    weight_precision_det: float = None

    def startup(self,
                ssm_scenario: NonLinearGaussian):
        init_cov_sqrt = ssm_scenario.initial_covariance_sqrt
        init_prec_sqrt = ssm_scenario.initial_precision_sqrt
        transition_cov_sqrt = ssm_scenario.transition_covariance_sqrt
        lik_mat = ssm_scenario.likelihood_matrix
        lik_cov_sqrt = ssm_scenario.likelihood_covariance_sqrt
        lik_prec_sqrt = ssm_scenario.likelihood_precision_sqrt

        self.initial_kalman_gain = kalman_gain(init_cov_sqrt, lik_mat, lik_cov_sqrt)
        lik_mat_t_lik_prec_sqrt = lik_mat.T @ lik_prec_sqrt
        init_cond_prec = init_prec_sqrt @ init_prec_sqrt \
                         + lik_mat_t_lik_prec_sqrt @ lik_mat_t_lik_prec_sqrt.T
        self.initial_conditioned_precision_sqrt = jnp.linalg.cholesky(init_cond_prec)
        self.initial_conditioned_covariance_sqrt = jnp.linalg.inv(self.initial_conditioned_precision_sqrt)
        self.initial_conditioned_precision_det = jnp.linalg.det(init_cond_prec)

        lik_mat_transition_cov_sqrt = lik_mat @ transition_cov_sqrt
        weight_precision = jnp.linalg.inv(lik_mat_transition_cov_sqrt @ lik_mat_transition_cov_sqrt.T
                                          + lik_cov_sqrt @ lik_cov_sqrt.T)
        self.proposal_kalman_gain = kalman_gain(transition_cov_sqrt, lik_mat, lik_cov_sqrt, inv_mat=weight_precision)

        proposal_cov = ssm_scenario.transition_covariance \
                       - self.proposal_kalman_gain @ ssm_scenario.likelihood_matrix \
                       @ ssm_scenario.transition_covariance
        self.proposal_covariance_sqrt = jnp.linalg.cholesky(proposal_cov)
        self.proposal_precision_sqrt = jnp.linalg.inv(self.proposal_covariance_sqrt)
        self.proposal_precision_det = 1 / jnp.linalg.det(proposal_cov)

        self.weight_precision_sqrt = jnp.linalg.cholesky(weight_precision)
        self.weight_precision_det = jnp.linalg.det(weight_precision)

    def initial_potential(self,
                          ssm_scenario: NonLinearGaussian,
                          x: jnp.ndarray,
                          y: jnp.ndarray,
                          t: float) -> Union[float, jnp.ndarray]:
        initial_conditioned_mean = ssm_scenario.initial_mean \
                                   + self.initial_kalman_gain \
                                   @ (y - ssm_scenario.likelihood_matrix @ ssm_scenario.initial_mean)
        return gaussian_potential(x, initial_conditioned_mean,
                                  sqrt_prec=self.initial_conditioned_precision_sqrt,
                                  det_prec=self.initial_conditioned_precision_det)

    def initial_sample(self,
                       ssm_scenario: NonLinearGaussian,
                       y: jnp.ndarray,
                       t: float,
                       random_key: jnp.ndarray) -> jnp.ndarray:
        initial_conditioned_mean = ssm_scenario.initial_mean \
                                   + self.initial_kalman_gain \
                                   @ (y - ssm_scenario.likelihood_matrix @ ssm_scenario.initial_mean)
        rand_shape = (ssm_scenario.dim,) if random_key.ndim == 1 else (len(random_key), ssm_scenario.dim)
        random_key = random_key if random_key.ndim == 1 else random_key[0]
        return initial_conditioned_mean \
               + random.normal(random_key, shape=rand_shape) @ self.initial_conditioned_covariance_sqrt

    def initial_log_weight(self,
                           ssm_scenario: NonLinearGaussian,
                           x: jnp.ndarray,
                           y: jnp.ndarray,
                           t: float) -> Union[float, jnp.ndarray]:
        return jnp.zeros(len(x))

    def initial_sample_and_weight_vectorised(self,
                                             ssm_scenario: NonLinearGaussian,
                                             y: jnp.ndarray,
                                             t: float,
                                             random_keys: jnp.ndarray) -> Tuple[jnp.ndarray, jnp.ndarray]:
        init_vals = self.initial_sample(ssm_scenario, y, t, random_keys)
        return init_vals, jnp.zeros(len(init_vals))

    def proposal_potential(self,
                           ssm_scenario: NonLinearGaussian,
                           x_previous: jnp.ndarray,
                           t_previous: float,
                           x_new: jnp.ndarray,
                           y_new: jnp.ndarray,
                           t_new: float) -> Union[float, jnp.ndarray]:
        mx = ssm_scenario.transition_function(x_previous, t_previous, t_new)
        conditioned_mean = mx + (y_new - mx @ ssm_scenario.likelihood_matrix.T) @ self.proposal_kalman_gain.T
        return gaussian_potential(x_new, conditioned_mean,
                                  sqrt_prec=self.proposal_precision_sqrt,
                                  det_prec=self.proposal_precision_det)

    def proposal_sample(self,
                        ssm_scenario: NonLinearGaussian,
                        x_previous: jnp.ndarray,
                        t_previous: float,
                        y_new: jnp.ndarray,
                        t_new: float,
                        random_key: jnp.ndarray) -> jnp.ndarray:
        mx = ssm_scenario.transition_function(x_previous, t_previous, t_new)
        conditioned_mean = mx + (y_new - mx @ ssm_scenario.likelihood_matrix.T) @ self.proposal_kalman_gain.T
        return conditioned_mean \
               + random.normal(random_key, shape=(ssm_scenario.dim,)) @ self.proposal_covariance_sqrt.T

    def intermediate_log_weight(self,
                                ssm_scenario: NonLinearGaussian,
                                x_previous: jnp.ndarray,
                                t_previous: float,
                                x_new: jnp.ndarray,
                                y_new: jnp.ndarray,
                                t_new: float) -> Union[float, jnp.ndarray]:
        mx = ssm_scenario.transition_function(x_previous, t_previous, t_new)
        return -gaussian_potential(y_new,
                                   mx @ ssm_scenario.likelihood_matrix.T,
                                   sqrt_prec=self.weight_precision_sqrt,
                                   det_prec=self.weight_precision_det)

    @partial(jit, static_argnums=(0, 1))
    def propose_and_intermediate_weight_vectorised(self,
                                                   ssm_scenario: NonLinearGaussian,
                                                   x_previous: jnp.ndarray,
                                                   t_previous: float,
                                                   y_new: jnp.ndarray,
                                                   t_new: float,
                                                   random_keys: jnp.ndarray) -> Tuple[jnp.ndarray, jnp.ndarray]:
        mx = vmap(ssm_scenario.transition_function, (0, None, None))(x_previous, t_previous, t_new)
        conditioned_mean = mx + (y_new - mx @ ssm_scenario.likelihood_matrix.T) @ self.proposal_kalman_gain.T
        x_new = conditioned_mean \
                + random.normal(random_keys[0], shape=x_previous.shape) @ self.proposal_covariance_sqrt.T

        log_weight_new = -gaussian_potential(y_new,
                                             mx @ ssm_scenario.likelihood_matrix.T,
                                             sqrt_prec=self.weight_precision_sqrt,
                                             det_prec=self.weight_precision_det)
        return x_new, log_weight_new


class EnsembleKalmanFilter(ParticleFilter):
    # NOTE we assume time-homogenous transition noise covariance
    # that is x_t ~ N(f_t(x_s), Q) where Q is independent of s, t or t-s
    # time-inhomogoneous transition noise covariance would require matrix inversions
    # at every time point rather than just once
    # typically this means t is of the form jnp.arange(start, end, step)

    name = 'Optimal Non-linear Gaussian Particle Filter'

    initial_kalman_gain: jnp.ndarray = None
    initial_conditioned_precision_sqrt: jnp.ndarray = None
    initial_conditioned_covariance_sqrt: jnp.ndarray = None

    def startup(self,
                ssm_scenario: NonLinearGaussian):
        init_cov_sqrt = ssm_scenario.initial_covariance_sqrt
        init_prec_sqrt = ssm_scenario.initial_precision_sqrt
        lik_mat = ssm_scenario.likelihood_matrix
        lik_cov_sqrt = ssm_scenario.likelihood_covariance_sqrt
        lik_prec_sqrt = ssm_scenario.likelihood_precision_sqrt

        self.initial_kalman_gain = kalman_gain(init_cov_sqrt, lik_mat, lik_cov_sqrt)
        lik_mat_t_lik_prec_sqrt = lik_mat.T @ lik_prec_sqrt
        init_cond_prec = init_prec_sqrt @ init_prec_sqrt \
                         + lik_mat_t_lik_prec_sqrt @ lik_mat_t_lik_prec_sqrt.T
        initial_conditioned_precision_sqrt = jnp.linalg.cholesky(init_cond_prec)
        self.initial_conditioned_covariance_sqrt = jnp.linalg.inv(initial_conditioned_precision_sqrt)

    def initial_sample_and_weight_vectorised(self,
                                             ssm_scenario: NonLinearGaussian,
                                             y: jnp.ndarray,
                                             t: float,
                                             random_keys: jnp.ndarray) -> Tuple[jnp.ndarray, jnp.ndarray]:
        initial_conditioned_mean = ssm_scenario.initial_mean \
                                   + self.initial_kalman_gain \
                                   @ (y - ssm_scenario.likelihood_matrix @ ssm_scenario.initial_mean)
        rand_shape = (len(random_keys), ssm_scenario.dim)
        init_x = initial_conditioned_mean \
                 + random.normal(random_keys[0], shape=rand_shape) @ self.initial_conditioned_covariance_sqrt
        return init_x, jnp.zeros(len(random_keys))

    @partial(jit, static_argnums=(0, 1))
    def propose_and_intermediate_weight_vectorised(self,
                                                   ssm_scenario: NonLinearGaussian,
                                                   x_previous: jnp.ndarray,
                                                   t_previous: float,
                                                   y_new: jnp.ndarray,
                                                   t_new: float,
                                                   random_keys: jnp.ndarray) -> Tuple[jnp.ndarray, jnp.ndarray]:
        n = x_previous.shape[0]
        split_keys = random.split(random_keys[0], 2)

        mx = vmap(ssm_scenario.transition_function, (0, None, None))(x_previous, t_previous, t_new) \
             + random.normal(split_keys[0], shape=x_previous.shape) @ ssm_scenario.transition_covariance_sqrt.T

        spread_matrix = (mx - mx.mean(0)).T / jnp.sqrt(n - 1)

        prop_kalman_gain = kalman_gain(spread_matrix,
                                       ssm_scenario.likelihood_matrix,
                                       ssm_scenario.likelihood_covariance_sqrt)

        y_prop = mx @ ssm_scenario.likelihood_matrix.T \
                 + random.normal(split_keys[1], shape=(n, len(y_new))) @ ssm_scenario.likelihood_covariance_sqrt.T

        x_new = mx + (y_new - y_prop) @ prop_kalman_gain.T

        return x_new, jnp.zeros(n)



########################################################################################################################
# Module: abc/scenarios/gk.py
# Description: Ricker model.
#
# Web: https://github.com/SamDuffield/mocat
########################################################################################################################

from typing import Tuple

from jax import numpy as jnp, random
from jax.lax import scan

from mocat.src.abc.abc import ABCScenario


def ricker_simulate(initial_n: int,
                    r: float,
                    noise_sd: float,
                    num_steps: int,
                    random_key: jnp.ndarray) -> jnp.ndarray:
    random_keys = random.split(random_key, num_steps)

    def body_fun(n: int, i: int) -> Tuple[int, int]:
        new_n = r * n * jnp.exp(-n + noise_sd * random.normal(random_keys[i]))
        return new_n, new_n

    _, ns = scan(body_fun, initial_n, jnp.arange(num_steps))
    return jnp.append(initial_n, ns)


class Ricker(ABCScenario):
    name: str = 'Ricker'
    dim: int = 3

    initial_n: int = 1
    num_steps: int = None
    observation_inds: jnp.ndarray = None
    data: jnp.ndarray = None

    def likelihood_sample(self,
                          x: jnp.ndarray,
                          random_key: jnp.ndarray) -> jnp.ndarray:
        dynamics_key, obs_key = random.split(random_key)
        ns = ricker_simulate(self.initial_n, x[0], x[1], self.num_steps, dynamics_key)
        observed_ns = ns[self.observation_inds]
        obs = random.poisson(obs_key, x[2] * observed_ns, shape=self.observation_inds.shape)
        return jnp.where(obs < 0, 0, obs)





########################################################################################################################
# Module: abc/scenarios/gk.py
# Description: G and K likelihood.
#
# Web: https://github.com/SamDuffield/mocat
########################################################################################################################

from abc import ABC as AbsBaseClass
from typing import Union

from jax import numpy as jnp, random, vmap
from jax.scipy.stats import norm

from mocat.src.abc.abc import ABCScenario


class _GK(ABCScenario):
    name: str = 'GK_fewN'
    n_unsummarised_data: int = None
    dim: int = 4

    c: float = 0.8

    prior_mins: Union[float, jnp.ndarray] = 0
    prior_maxs: Union[float, jnp.ndarray] = 10

    buffer: float = 1e-5

    def full_data_sample(self,
                         x: jnp.ndarray,
                         random_key: jnp.ndarray) -> jnp.ndarray:
        raise NotImplementedError(f'{self.name} full_data_sample not implemented')

    def summarise_data(self,
                       data: jnp.ndarray) -> jnp.ndarray:
        raise NotImplementedError(f'{self.name} summarise_data not implemented')

    def likelihood_sample(self,
                          x: jnp.ndarray,
                          random_key: jnp.ndarray) -> jnp.ndarray:
        return self.summarise_data(self.full_data_sample(x, random_key))


class GKUniformPrior(_GK, AbsBaseClass):

    def full_data_sample(self,
                         x: jnp.ndarray,
                         random_key: jnp.ndarray) -> jnp.ndarray:
        u = random.uniform(random_key, shape=(self.n_unsummarised_data,), minval=self.buffer, maxval=1 - self.buffer)
        z = norm.ppf(u)
        expmingz = jnp.exp(-x[2] * z)
        return x[0] \
               + x[1] * (1 + self.c * (1 - expmingz) / (1 + expmingz)) \
               * z * (1 + z ** 2) ** x[3]

    def prior_potential(self,
                        x: jnp.ndarray,
                        random_key: jnp.ndarray = None) -> Union[float, jnp.ndarray]:
        out = jnp.where(jnp.all(x > self.prior_mins), 1., jnp.inf)
        out = jnp.where(jnp.all(x < self.prior_maxs), out, jnp.inf)
        return out

    def prior_sample(self,
                     random_key: jnp.ndarray) -> Union[float, jnp.ndarray]:
        return self.prior_mins + random.uniform(random_key, (self.dim,)) * (self.prior_maxs - self.prior_mins)


class GKTransformedUniformPrior(_GK, AbsBaseClass):

    def constrain(self,
                  unconstrained_x: jnp.ndarray):
        return self.prior_mins + norm.cdf(unconstrained_x) * (self.prior_maxs - self.prior_mins)

    def unconstrain(self,
                    constrained_x: jnp.ndarray):
        return norm.ppf((constrained_x - self.prior_mins) / (self.prior_maxs - self.prior_mins))

    def full_data_sample(self,
                         x: jnp.ndarray,
                         random_key: jnp.ndarray) -> jnp.ndarray:
        transformed_x = self.constrain(x)
        u = random.uniform(random_key, shape=(self.n_unsummarised_data,), minval=self.buffer, maxval=1 - self.buffer)
        z = norm.ppf(u)
        expmingz = jnp.exp(-transformed_x[2] * z)
        return transformed_x[0] \
               + transformed_x[1] * (1 + self.c * (1 - expmingz) / (1 + expmingz)) \
               * z * (1 + z ** 2) ** transformed_x[3]

    def prior_potential(self,
                        x: jnp.ndarray,
                        random_key: jnp.ndarray = None) -> Union[float, jnp.ndarray]:
        return 0.5 * (x ** 2).sum()

    def prior_sample(self,
                     random_key: jnp.ndarray) -> Union[float, jnp.ndarray]:
        return random.normal(random_key, (self.dim,))


class GKOnlyAUniformPrior(_GK, AbsBaseClass):
    name: str = 'GK_fewN only A'
    dim: int = 1

    B: float = 1.
    g: float = 2.
    k: float = 0.5

    def full_data_sample(self,
                         x: jnp.ndarray,
                         random_key: jnp.ndarray) -> jnp.ndarray:
        u = random.uniform(random_key, shape=(self.n_unsummarised_data,), minval=self.buffer, maxval=1 - self.buffer)
        z = norm.ppf(u)
        expmingz = jnp.exp(-self.g * z)
        return x[0] \
               + self.B * (1 + self.c * (1 - expmingz) / (1 + expmingz)) \
               * z * (1 + z ** 2) ** self.c

    def prior_potential(self,
                        x: jnp.ndarray,
                        random_key: jnp.ndarray = None) -> Union[float, jnp.ndarray]:
        out = jnp.where(jnp.all(x > self.prior_mins), 1., jnp.inf)
        out = jnp.where(jnp.all(x < self.prior_maxs), out, jnp.inf)
        return out

    def prior_sample(self,
                     random_key: jnp.ndarray) -> Union[float, jnp.ndarray]:
        return self.prior_mins + random.uniform(random_key, (self.dim,)) * (self.prior_maxs - self.prior_mins)


class GKOnlyATransformedUniformPrior(_GK, AbsBaseClass):
    name: str = 'GK_fewN only A'
    dim: int = 1

    B: float = 1.
    g: float = 2.
    k: float = 0.5

    def constrain(self,
                  unconstrained_x: jnp.ndarray):
        return self.prior_mins + norm.cdf(unconstrained_x) * (self.prior_maxs - self.prior_mins)

    def unconstrain(self,
                    constrained_x: jnp.ndarray):
        return norm.ppf((constrained_x - self.prior_mins) / (self.prior_maxs - self.prior_mins))

    def full_data_sample(self,
                         x: jnp.ndarray,
                         random_key: jnp.ndarray) -> jnp.ndarray:
        transformed_x = self.constrain(x)
        u = random.uniform(random_key, shape=(self.n_unsummarised_data,), minval=self.buffer, maxval=1 - self.buffer)
        z = norm.ppf(u)
        expmingz = jnp.exp(-self.g * z)
        return transformed_x[0] \
               + self.B * (1 + self.c * (1 - expmingz) / (1 + expmingz)) \
               * z * (1 + z ** 2) ** self.c

    def prior_potential(self,
                        x: jnp.ndarray,
                        random_key: jnp.ndarray = None) -> Union[float, jnp.ndarray]:
        return 0.5 * (x ** 2).sum()

    def prior_sample(self,
                     random_key: jnp.ndarray) -> Union[float, jnp.ndarray]:
        return random.normal(random_key, (self.dim,))
